"""
Bluetooth Low Energy Peripheral abstraction
===========================================

This module provides the PeripheralDevice class used to wrap all GATT operations
for a given connected device. This class wraps all the following operations:

* service and characteristics discovery
* ATT MTU exchange
* characteristic and descriptor read
* characteristic and descriptor write

For instance, a `Central` object can be used to initiate a BLE connection to a
target, and will return a `PeripheralDevice` object, as shown below::

    central = Central(...)
    target = central.connect('00:11:22:33:44:55')

One can then use this object to discover all the services and characteristics::

    target.discover()

And look for a specific characteristic and read it::

    device_name = target.get_characteristic(UUID('1800'), UUID('2A00'))
    if device_name is not None:
        print('Device name is {}'.format(device_name.read()))

It is also possible to write to a characteristic (if writeable)::

    device_name.value = b'MyNewDeviceName'

"""
import logging

from whad.ble.profile.service import Service
from whad.ble.profile.characteristic import CharacteristicDescriptor, \
    CharacteristicProperties, Characteristic, CharacteristicValue
from whad.ble.profile import GenericProfile
from whad.ble.stack.att.constants import BleAttProperties
from whad.ble.profile.attribute import UUID

from struct import unpack
from time import sleep

logger = logging.getLogger(__name__)

class PeripheralCharacteristicDescriptor:
    """Wrapper for a peripheral characteristic descriptor.
    """

    def __init__(self, descriptor, gatt):
        """Initialize a PeripheralCharacteristicDescriptor.

        :param CharacteristicDescriptor descriptor: Descriptor to wrap.
        :param GattClient gatt: GATT client to use for GATT operations (read/write).
        """
        self.__descriptor = descriptor
        self.__gatt = gatt

    @property
    def handle(self):
        """Return this characteristic descriptor handle.

        :return int: Descriptor handle
        """
        return self.__descriptor.handle

    @property
    def type_uuid(self):
        """Return this attribute type UUID.

        :return UUID: Attribute type UUID
        """
        return self.__descriptor.type_uuid

    def read(self):
        """Read descriptor value.

        :return bytes: Descriptor value
        """
        return self.__gatt.read(self.__descriptor.handle)

    def write(self, value, without_response=False):
        """Write descriptor value.

        :param bytes value: Value to write to this descriptor.
        :param bool without_response: If set, use a GATT write command request to write to this descriptor.
        """
        if without_response:
            self.__gatt.write_command(self.__descriptor.handle, value)
        else:
            self.__gatt.write(self.__descriptor.handle, value)

class PeripheralCharacteristicValue:
    """CharacteristicValue wrapper for peripheral devices

    Forward all read/write operations to PeripheralCharacteristic wrapper
    because initially access to characteristic has been implemented there.

    Could be interesting in the future to implement all these operations here
    and to forward from characteristic wrapper to characteristic value wrapper
    because it makes more sense.

    Anyway, this code just works but from an architectural point-of-view is
    a bit crappy.
    """

    def __init__(self, char_value, gatt):
        self.__char_value = char_value
        self.__characteristic = PeripheralCharacteristic(
            char_value.characteristic,
            gatt
        )
        self.__gatt = gatt

    @property
    def handle(self):
        return self.__char_value.handle

    @property
    def characteristic(self):
        return self.__characteristic

    @property
    def value(self):
        """Transparent characteristic read.

        :return bytes: Characteristic value
        """
        return self.__characteristic.read()

    @value.setter
    def value(self, val):
        """Transparent characteristic write.

        :param bytes val: Value to write into this characteristic
        """
        return self.__characteristic.write(val)

    def read(self, offset=0):
        """Read characteristic value
        """
        return self.__characteristic.read(offset=offset)

    def read_long(self):
        return self.__characteristic.read_long()

    def write(self, value, without_response=False):
        return self.__characteristic.write(value, without_response=without_response)


class PeripheralCharacteristic:
    """Characteristic wrapper for peripheral devices

    Instruments gatt to read/write a remote characteristic.
    """
    def __init__(self, characteristic, gatt):
        self.__characteristic = characteristic
        self.__gatt = gatt

    @property
    def value(self):
        """Transparent characteristic read.

        :return bytes: Characteristic value
        """
        return self.read()

    @value.setter
    def value(self, val):
        """Transparent characteristic write.

        :param bytes val: Value to write into this characteristic
        """
        return self.write(val)

    @property
    def uuid(self):
        return self.__characteristic.uuid

    @property
    def type_uuid(self):
        return self.__characteristic.type_uuid

    @property
    def properties(self):
        return self.__characteristic.properties

    @property
    def handle(self):
        return self.__characteristic.handle

    @property
    def end_handle(self):
        return self.__characteristic.end_handle

    @property
    def value_handle(self):
        return self.__characteristic.value_handle

    def can_notify(self):
        return self.__characteristic.can_notify()

    def must_notify(self):
        return self.__characteristic.must_notify()

    def can_indicate(self):
        return self.__characteristic.can_indicate()

    def must_indicate(self):
        return self.__characteristic.must_indicate()

    def read(self, offset=0):
        """Read characteristic value
        """
        if offset == 0:
            return self.__gatt.read(self.__characteristic.value_handle)
        else:
            return self.__gatt.read_blob(self.__characteristic.value_handle, offset)

    def read_long(self):
        """Read long characteristic value
        """
        return self.__gatt.read_long(self.__characteristic.value_handle)

    def write(self, value, without_response=False):
        """Set characteristic value

        If characteristic is only writeable without response, use a write command
        rather than a write request. Otherwise, use a write request. If a characteristic
        has both write and write without response properties, `without_response` must be
        set to True to use a write command.
        """
        # If characteristic is only writeable without response, force without_response to True.
        access_mask = CharacteristicProperties.WRITE_WITHOUT_RESPONSE | CharacteristicProperties.WRITE
        if (self.__characteristic.properties & access_mask) == CharacteristicProperties.WRITE_WITHOUT_RESPONSE:
            without_response = True

        if isinstance(value, bytes):
            if without_response:
                return self.__gatt.write_command(
                    self.__characteristic.value_handle,
                    value
                )
            else:
                return self.__gatt.write(
                    self.__characteristic.value_handle,
                    value
                )

    def descriptors(self):
        """Return all the descriptors associated with this characteristic.

        This method is a generator and will yield all the descriptors registered
        with this characteristic.
        """
        for desc in self.__characteristic.descriptors():
            yield PeripheralCharacteristicDescriptor(
                desc,
                self.__gatt
            )

    def get_descriptor(self, type_uuid):
        """Get descriptor of a given type.

        :param UUID type_uuid: Descriptor type UUID
        :return PeripheralCharacteristicDescriptor: Return descriptor if found, `None` otherwise
        """
        for desc in self.__characteristic.descriptors():
            if desc.type_uuid == type_uuid:
                return PeripheralCharacteristicDescriptor(
                    desc,
                    self.__gatt
                )

    def readable(self):
        """Check if this characteristic is readable.

        :return bool: True if readable, False otherwise.
        """
        return ((self.__characteristic.properties & CharacteristicProperties.READ) != 0)

    def writeable(self):
        """Check if this characteristic is writeable.

        :return bool: True if writeable, False otherwise.
        """
        return (
            ((self.__characteristic.properties & CharacteristicProperties.WRITE) != 0) or
            ((self.__characteristic.properties & CharacteristicProperties.WRITE_WITHOUT_RESPONSE) != 0)
        )

    def subscribe(self, notification=True, indication=False, callback=None):
        """Subscribe for notification/indication.

        :param bool notification: If set, subscribe for notification
        :param bool indication: If set, subscribe for indication (cannot be used when notification is set)
        :param callable callback: Callback function to be called on indication/notification event
        :return bool: True if subscription has successfully been performed, False otherwise.
        """
        if notification:
            # Look for CCCD
            desc = self.get_descriptor(UUID(0x2902))
            if desc is not None:
                # wrap our callback to provide more details about the concerned
                # characteristic
                def wrapped_cb(handle, value, indication=False):
                    callback(
                        self,
                        value,
                        indication=indication
                    )

                # Register our callback
                if callback is not None:
                    self.__gatt.register_notification_callback(
                        self.__characteristic.value_handle,
                        wrapped_cb
                    )

                # Enable notification
                desc.write(bytes([0x01, 0x00]))

                return True
            else:
                return False
        elif indication:
            # Look for CCCD
            desc = self.get_descriptor(UUID(0x2902))
            if desc is not None:
                # Register our callback
                if callback is not None:
                    self.__gatt.register_notification_callback(
                        self.__characteristic.value_handle,
                        callback
                    )

                # Enable indication
                desc.write(bytes([0x02, 0x00]))

                return True
            else:
                return False

    def unsubscribe(self):
        """Unsubscribe from this characteristic.
        """
        # Look for CCCD
        desc = self.get_descriptor(UUID(0x2902))

        if desc is not None:
            # Disable notification/indication
            desc.write(bytes([0x00, 0x00]))

            # Unregister our callback
            self.__gatt.unregister_notification_callback(
                self.__characteristic.value_handle
            )
            return True
        else:
            return False


class PeripheralService:
    """Service wrapper for peripheral devices
    """

    def __init__(self, service, gatt):
        self.__service = service
        self.__gatt = gatt

    @property
    def handle(self):
        """Return this service handle.

        :param int: service handle
        """
        return self.__service.handle

    @property
    def end_handle(self):
        """Return this service end handle.

        :return int: end handle
        """
        return self.__service.end_handle

    @property
    def uuid(self):
        """Return this service UUID.

        :param UUID: Service UUID
        """
        return self.__service.uuid

    @property
    def type_uuid(self):
        """Return this service type UUID

        :return UUID: Service type UUID
        """
        return self.__service.type_uuid

    def read_characteristic_by_uuid(self, uuid):
        """Read a characteristic belonging to this service identified by its UUID.

        :param UUID uuid: Characteristic UUID
        :return bytes: Characteristic value
        """
        return self.__gatt.read_characteristic_by_uuid(
            uuid,
            self.__service.handle,
            self.__service.end_handle
        )

    def get_characteristic(self, uuid):
        """Look for a specific characteristic belonging to this service, identified by its UUID.

        :param UUID uuid: Characteristic UUID
        :return PeripheralCharacteristic: Found characteristic if any, `None` otherwise.
        """
        for charac in self.__service.characteristics():
            if charac.uuid == uuid:
                return PeripheralCharacteristic(
                    charac,
                    self.__gatt
                )
        return None

    def characteristics(self):
        """Enumerate this service's characteristics (generator).
        """
        for characteristic in self.__service.characteristics():
            yield PeripheralCharacteristic(
                characteristic,
                self.__gatt
            )

class PeripheralDevice(GenericProfile):
    """GATT client wrapper representing a remote device.

    This class is used to wrap a device model used in a gatt client
    in order to provide easy-to-use methods to access its services,
    characteristics and descriptors.
    """

    def __init__(self,  central, gatt_client, conn_handle, from_json=None):
        """Create a peripheral device from a Central and a GATT client.

        :param  central:        Central instance used to connect to a target device.
        :type   central:        :class:`whad.ble.connector.central.Central`
        :param  gatt_client:    GATT client connected to a target device.
        :type   gatt_client:    :class:`whad.ble.stack.gatt.GattClient`
        :param  conn_handle:    Current connection handle.
        :type   conn_handle:    int
        :param  from_json:      GATT profile (JSON) to be used when instanciating the underlying GattProfile.
        :type   from_json:      str, optional
        """
        self.__gatt = gatt_client
        self.__smp = gatt_client.smp
        self.__ll = gatt_client.get_layer('ll')
        self.__conn_handle = conn_handle
        self.__central = central
        self.__disconnect_cb = None
        super().__init__(from_json=from_json)


    @property
    def conn_handle(self) -> int:
        """Current connection handle.
        """
        return self.__conn_handle


    def start_encryption(self):
        security_database = self.__smp.security_database


        crypto_material = security_database.get(address=self.__central.target_peer)
        conn_handle = self.__smp.get_layer('l2cap').state.conn_handle
        if crypto_material is not None and crypto_material.has_ltk():
            self.__ll.start_encryption(
                conn_handle,
                unpack('>Q', crypto_material.ltk.rand)[0],
                crypto_material.ltk.ediv
            )

    def pairing(self, pairing=None):
        """Trigger a pairing according to provided parameters.
        Default parameters will be used if pairing parameter is None.
        """
        if not self.__smp.initiate_pairing(parameters=pairing):
            return False

        while not self.__smp.is_pairing_done():
            sleep(0.1)
            if self.__smp.is_pairing_failed():
                return False

        self.__smp.reset_state()
        return True

    def set_disconnect_cb(self, callback):
        """Set disconnection callback.

        :param callback:    Callback function to call on disconnection.
        :type   callback:   callable
        """
        self.__disconnect_cb = callback


    def set_mtu(self, mtu: int):
        """Update connection MTU.

        :param  mtu:    ATT MTU to use for this connection.
        :type   mtu:    int
        :return:        Remote device MTU.
        :rtype: int
        """
        return self.__gatt.set_mtu(mtu)


    def disconnect(self):
        """Terminate the connection to this device
        """
        # Ask associated central to disconnect this peripheral device.
        self.__central.disconnect(self.__conn_handle)


    def discover(self):
        """Discovers services, characteristics and descriptors.

        This method must be called before accessing any service or characteristic,
        as it is required to retrieve the corresponding GATT handles.
        """
        # Discover
        self.__gatt.discover()


    def find_service_by_uuid(self, uuid: UUID) -> PeripheralService:
        """Find service by its UUID

        :param  uuid:   Characteristic UUID
        :type   uuid:   :class:`whad.ble.profile.attribute.UUID`
        :return:        PeripheralService: An instance of PeripheralService if service has been found, None otherwise.
        :rtype: :class:`whad.ble.profile.device.PeripheralService`
        """
        service = self.__gatt.discover_primary_service_by_uuid(uuid)
        if service is not None:
            return PeripheralService(
                service,
                self.__gatt
            )
        else:
            return None

    def find_characteristic_by_uuid(self, uuid: UUID):
        """Find characteristic by its UUID

        :param  uuid:   Characteristic UUID
        :type   uuid:   :class:`whad.ble.profile.attribute.UUID`
        :return:        PeripheralCharacteristic: An instance of PeripheralCharacteristic if characteristic has been found, None otherwise.
        :rtype: :class:`whad.ble.profile.device.PeripheralCharacteristic`
        """
        for service in self.services():
            for charac in service.characteristics():
                if charac.uuid == uuid:
                    return PeripheralCharacteristic(
                        charac,
                        self.__gatt
                    )


    def find_object_by_handle(self, handle):
        """Find an existing object (service, attribute, descriptor) based on its handle,
        it known from the underlying GenericProfile.

        :param  handle: Object handle
        :type   handle: int
        :return:        Characteristic, characteristic value or service
        :rtype:         :class:`whad.ble.profile.device.PeripheralCharacteristic`, :class:`whad.ble.profile.device.PeripheralCharacteristicValue`, :class:`whad.ble.profile.device.PeripheralService`
        """
        obj = super().find_object_by_handle(handle)
        if isinstance(obj, Characteristic):
            return PeripheralCharacteristic(
                obj,
                self.__gatt
            )
        elif isinstance(obj, Service):
            return PeripheralService(
                obj,
                self.__gatt
            )
        elif isinstance(obj, CharacteristicValue):
            return PeripheralCharacteristicValue(
                obj,
                self.__gatt
            )

    def get_characteristic(self, service_uuid: UUID, charac_uuid: UUID):
        """Get a PeripheralCharacteristic object representing a characteristic
        defined by the given service UUID and characteristic UUID.

        :param  service_uuid:   Service UUID
        :type   service_uuid:   :class:`whad.ble.profile.attribute.UUID`
        :param  charac_uuid:    Characteristic UUID
        :type   charac_uuid:    :class:`whad.ble.profile.attribute.UUID`
        :return:                PeripheralCharacteristic object on success, None if not found.
        :rtype: :class:`whad.ble.profile.device.PeripheralCharacteristic`
        """
        service = self.get_service(service_uuid)
        if service is not None:
            return service.get_characteristic(charac_uuid)
        return None


    def get_service(self, uuid):
        """Retrieve a PeripheralService object given its UUID.

        :param  uuid:       Service UUID
        :type   uuid:       :class:`whad.ble.profile.attribute.UUID`
        :return:            Corresponding PeripheralService object if found, None otherwise.
        :rtype: :class:`whad.ble.profile.device.PeripheralService`
        """

        for service in self.services():
            if service.uuid == uuid:
                return PeripheralService(service, self.__gatt)
        return None


    def write(self, handle, value):
        """Perform a write operation on an attribute based on its handle.

        This method allows to interact with characteristics and descriptors without
        having performing a GATT services and characteristics discovery. One just need
        to specify the handle corresponding to a characteristic value or descriptor and the
        value to write to it, and our GATT stack will handle it.

        Note that there is absolutely no check on corresponding characteristic permissions
        (meaning you can try to write on a read-only characteristic value) and that this
        method may raise exceptions due to potential GATT errors the remote device may
        return.

        :param  handle: Characteristic or descriptor handle to write.
        :type   handle: int
        :param  value:  Bytes to write into this characteristic.
        :type   value:  bytes
        """
        return self.__gatt.write(handle, value)


    def write_command(self, handle, value, without_response=False):
        """Perform a write command operation (no write response will be sent) on
        an attribute based on its handle.

        This method allows to interact with characteristics and descriptors without
        having performing a GATT services and characteristics discovery. One just need
        to specify the handle corresponding to a characteristic value or descriptor and the
        value to write to it, and our GATT stack will handle it.

        Note that there is absolutely no check on corresponding characteristic permissions
        (meaning you can try to write on a read-only characteristic value) and that this
        method may raise exceptions due to potential GATT errors the remote device may
        return.

        :param  handle: Characteristic or descriptor handle to write.
        :type   handle: int
        :param  value:  Bytes to write into this characteristic.
        :type   value:  bytes
        """
        return self.__gatt.write_command(handle, value)

    def read(self, handle, offset=None, long=False):
        """Perform a read operation on an attribute based on its handle.

        This method allows to interact with characteristics and descriptors without
        having performing a GATT services and characteristics discovery. One just need
        to specify the handle corresponding to a characteristic value or descriptor
        and our GATT stack will handle it.

        Note that there is absolutely no check on corresponding characteristic permissions
        (meaning you can try to read from a write-only characteristic value) and that this
        method may raise exceptions due to potential GATT errors the remote device may
        return.

        :param  handle: Characteristic or descriptor handle.
        :type   handle: int
        :param  offset: Offset applied when reading data from characteristic or descriptor (default: 0).
        :type   offset: int, optional
        :param  long:   use GATT long read procedure if set to True (default: False)
        :type   long:   bool, optional
        :return:        Content of the characteristic or descriptor.
        :rtype: bytes
        """
        if not long:
            if offset is None:
                return self.__gatt.read(handle)
            else:
                return self.__gatt.read_blob(handle, offset=offset)
        else:
            return self.__gatt.read_long(handle)


    def on_disconnect(self, conn_handle):
        """Disconnection callback

        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        """
        logger.debug('PeripheralDevice has disconnected')
        if self.__disconnect_cb is not None:
            self.__disconnect_cb()



"""
Pythonic Enhanced ShockBurst stack
"""
from whad.esb.stack.llm import LinkLayer
from whad.common.stack import Layer, alias, source

import logging
logger = logging.getLogger(__name__)

@alias('phy')
class ESBStack(Layer):
    """
    This class holds the main components of a (hackable) Enhanced ShockBurst stack:

    - the Link Layer (LL)
    - the Applicative Layer (APP)

    The Link Layer handles all the low-level operations, e.g., packet transmission, reception, synchronization and acknowledgements.
    Generally, a proprietary Applicative layer is implemented on the top of this stack (e.g., Unifying, Microsoft).

    """

    def __init__(self, connector, options={}):
        """
        Create an instance of Enhanced ShockBurst associated with a specific connector. This
        connector provides the transport layer.

        :param WhadDeviceConnector connector: Connector to use with this stack.
        """

        super().__init__(options=options)

        # Save connector (used as PHY layer)
        self.__connector = connector

    @property
    def ll(self):
        return self.get_layer('ll')


    @property
    def app(self):
        return self.get_layer('app')

    def on_pdu(self, pdu):
        '''PDU callback.

        This callback handles a received PDU.
        '''
        logger.debug('received a PDU (%d bytes).' % (len(pdu)))
        self.send('ll', pdu)


    @property
    def channel(self):
        return self.__connector.channel

    @channel.setter
    def channel(self, channel):
        self.__connector.channel = channel

    @property
    def address(self):
        return self.__connector.address

    @address.setter
    def address(self, address):
        self.__connector.address = address

    @source('ll')
    def send_pdu(self, packet, channel=None, retransmission_count=1):
        if channel is None:
            channel = self.__connector.channel
        logger.debug('transmitted a PDU (%d bytes).' % (len(packet)))

        return self.__connector.send(packet, channel=channel, retransmission_count=retransmission_count)

ESBStack.add(LinkLayer)
'''
class ESBStack:
    def __init__(self, connector, app_class=None):
        """
        Create an instance of Enhanced ShockBurst associated with a specific connector. This
        connector provides the transport layer.

        :param WhadDeviceConnector connector: Connector to use with this stack.
        """
        self.__connector = connector

        # Instanciate all the required controllers
        self.__llm = EsbLinkLayerManager(self, app_class)

    @property
    def ll(self):
        return self.__llm

    @property
    def app(self):
        return self.__llm.app
    #############################
    # Incoming messages
    #############################
    def on_pdu(self, pdu):
        self.__llm.on_pdu(pdu)

    ############################
    # Interact
    ############################

    @property
    def channel(self):
        return self.__connector.channel

    @channel.setter
    def channel(self, channel):
        self.__connector.channel = channel

    @property
    def address(self):
        return self.__connector.address

    @address.setter
    def address(self, address):
        self.__connector.address = address

    def send(self, packet, channel=None, retransmission_count=1):
        if channel is None:
            channel = self.__connector.channel
        return self.__connector.send(packet, channel=channel, retransmission_count=retransmission_count)
'''



"""
Whad up ?
"""
import sys
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.device import WhadDevice
from whad.protocol.ble.ble_pb2 import BleCommand
from whad.protocol.dot15d4.dot15d4_pb2 import Dot15d4Command
from whad.protocol.esb.esb_pb2 import ESBCommand
from whad.protocol.unifying.unifying_pb2 import UnifyingCommand
from whad.protocol.phy.phy_pb2 import PhyCommand
from whad.hub.discovery import Domain, Capability
from prompt_toolkit import print_formatted_text, HTML
from html import escape

import logging
#logging.basicConfig(level=logging.DEBUG)

DOMAINS = {
    Domain.Phy: 'Physical Layer',
    Domain.ANT: 'ANT',
    Domain.ANT_FS: 'ANT FS',
    Domain.ANT_Plus: 'ANT+',
    Domain.BtClassic: 'Bluetooth Classic',
    Domain.BtLE: 'Bluetooth LE',
    Domain.Esb: 'Enhanced ShockBurst',
    Domain.LogitechUnifying: 'Logitech Unifying',
    Domain.Mosart: 'Mosart',
    Domain.SixLowPan: '6LowPan',
    Domain.Dot15d4: '802.15.4'
}

CAPABILITIES = {
    Capability.Scan: 'can scan devices',
    Capability.Hijack: 'can hijack communication',
    Capability.Hook: 'can hook packets',
    Capability.Inject: 'can inject packets',
    Capability.Jam: 'can jam communications',
    Capability.SimulateRole: 'can simulate a role in a communication',
    Capability.Sniff: 'can sniff data',
    Capability.NoRawData: 'can not read/write raw packet'
}

BLE_COMMANDS = {
    BleCommand.SetBdAddress: 'SetBdAddress: can set BD address',
    BleCommand.SniffAdv: 'SniffAdv: can sniff advertising PDUs',
    BleCommand.JamAdv: 'JamAdv: can jam advertising PDUs',
    BleCommand.JamAdvOnChannel: 'JamAdvOnChannel: can jam advertising PDUs on a single channel',
    BleCommand.ReactiveJam: 'ReactiveJam: can reactively jam PDU on a single channel',
    BleCommand.SniffConnReq: 'SniffConnReq: can sniff a new connection',
    BleCommand.SniffAccessAddress: 'SniffAccessAddress: can detect active connections',
    BleCommand.SniffActiveConn: 'SniffActiveConn: can sniff an active connection',
    BleCommand.JamConn: 'JamConn: can jam an active connection',
    BleCommand.ScanMode: 'ScanMode: can scan devices',
    BleCommand.AdvMode: 'AdvMode: can advertise as a BLE device',
    BleCommand.SetAdvData: 'SetAdvData: can set advertising PDU details',
    BleCommand.CentralMode: 'CentralMode: can act as a Central device',
    BleCommand.ConnectTo: 'ConnectTo: can initiate a BLE connection',
    BleCommand.SendRawPDU: 'SendRawPDU: can send a raw PDU',
    BleCommand.SendPDU: 'SendPDU: can send a PDU',
    BleCommand.Disconnect: 'Disconnect: can terminate an active connection (in Central mode)',
    BleCommand.PeripheralMode: 'PeripheralMode: can act as a peripheral',
    BleCommand.Start: 'Start: can start depending on the current mode',
    BleCommand.Stop: 'Stop: can stop depending on the current mode',
    BleCommand.SetEncryption: 'SetEncryption: can enable encryption during a connection',
    BleCommand.HijackMaster: 'HijackMaster: can hijack the Master role in an active connection',
    BleCommand.HijackSlave: 'HijackSlave: can hijack the Slave role in an active connection',
    BleCommand.HijackBoth: 'HijackBoth: can hijack the Master and the Slave role in an active connection',
    BleCommand.PrepareSequence: 'PrepareSequence: can prepare a sequence of packets and associate a trigger',
    BleCommand.TriggerSequence: 'TriggerSequence: can manually trigger the transmission of a sequence of packets',
    BleCommand.DeleteSequence: 'DeleteSequence: can delete a prepared sequence of packets'

}

DOT15D4_COMMANDS = {
    Dot15d4Command.SetNodeAddress: "SetNodeAddress: can set Node address",
    Dot15d4Command.Sniff: "Sniff: can sniff 802.15.4 packets",
    Dot15d4Command.EnergyDetection: "EnergyDetection: can perform energy detection scans",
    Dot15d4Command.Jam: "Jam: can jam 802.15.4 packets",
    Dot15d4Command.Send: "Send: can transmit 802.15.4 packets",
    Dot15d4Command.EndDeviceMode: "EndDeviceMode: can act as an End Device",
    Dot15d4Command.CoordinatorMode: "CoordinatorMode: can act as a Coordinator",
    Dot15d4Command.RouterMode: "RouterMode: can act as a Router",
    Dot15d4Command.Start: "Start: can start depending on the current mode",
    Dot15d4Command.Stop: "Stop: can stop depending on the current mode",
    Dot15d4Command.ManInTheMiddle: "ManInTheMiddle: can perform a Man-in-the-Middle attack",
}

ESB_COMMANDS = {
    ESBCommand.SetNodeAddress: "SetNodeAddress: can set Node address",
    ESBCommand.Sniff: "Sniff: can sniff Enhanced ShockBurst packets",
    ESBCommand.Jam: "Jam: can jam Enhanced ShockBurst packets",
    ESBCommand.Send: "Send: can transmit Enhanced ShockBurst packets",
    ESBCommand.PrimaryReceiverMode: "PrimaryReceiverMode: can act as a Primary Receiver (PRX)",
    ESBCommand.PrimaryTransmitterMode: "PrimaryReceiverMode: can act as a Primary Receiver (PTX)",
    ESBCommand.Start: "Start: can start depending on the current mode",
    ESBCommand.Stop: "Stop: can stop depending on the current mode"
}

UNIFYING_COMMANDS = {
    UnifyingCommand.SetNodeAddress: "SetNodeAddress: can set Node address",
    UnifyingCommand.Sniff: "Sniff: can sniff Logitech Unifying packets",
    UnifyingCommand.Jam: "Jam: can jam Logitech Unifying packets",
    UnifyingCommand.Send: "Send: can transmit Logitech Unifying packets",
    UnifyingCommand.LogitechDongleMode: "PrimaryReceiverMode: can act as a Logitech Dongle (ESB PRX)",
    UnifyingCommand.LogitechKeyboardMode: "PrimaryReceiverMode: can act as a Logitech Keyboard (ESB PTX)",
    UnifyingCommand.LogitechMouseMode: "LogitechMouseMode: can act as a Logitech Mouse (ESB PTX)",
    UnifyingCommand.Start: "Start: can start depending on the current mode",
    UnifyingCommand.Stop: "Stop: can stop depending on the current mode",
    UnifyingCommand.SniffPairing: "SniffPairing: can sniff a pairing process"
}

PHY_COMMANDS = {
    PhyCommand.SetASKModulation: "SetASKModulation: can use Amplitude Shift Keying modulation scheme",
    PhyCommand.SetFSKModulation: "SetFSKModulation: can use Frequency Shift Keying modulation scheme",
    PhyCommand.SetGFSKModulation: "SetGFSKModulation: can use Gaussian Frequency Shift Keying modulation scheme",
    PhyCommand.SetBPSKModulation: "SetBPSKModulation: can use Binary Phase Shift Keying modulation scheme",
    PhyCommand.SetQPSKModulation: "SetQPSKModulation: can use Quadrature Phase Shift Keying modulation scheme",
    PhyCommand.GetSupportedFrequencies: "GetSupportedFrequencies: can return a list of supported frequency ranges",
    PhyCommand.SetFrequency: "SetFrequency: can configure a given frequency",
    PhyCommand.SetDataRate: "SetDataRate: can configure the datarate",
    PhyCommand.SetEndianness: "SetEndianness: can configure the endianness",
    PhyCommand.SetTXPower: "SetTXPower: can configure the transmission power level",
    PhyCommand.SetPacketSize: "SetPacketSize: can configure the packet size",
    PhyCommand.SetSyncWord: "SetSyncWord: can configure the synchronization word",
    PhyCommand.Sniff: "Sniff: can receive arbitrary packets",
    PhyCommand.Send: "Send: can transmit arbitrary packets",
    PhyCommand.SendRaw: "SendRaw: can transmit arbitrary IQ streams" ,
    PhyCommand.ScheduleSend: "ScheduleSend: can schedule a packet to be sent at a specific time",
    PhyCommand.Jam: "Jam: can jam a physical medium",
    PhyCommand.Monitor: "Monitor: can monitor a physical medium",
    PhyCommand.Start: "Start: can start depending on the current mode",
    PhyCommand.Stop: "Stop: can stop depending on the current mode",
}
COMMANDS = {
    Domain.BtLE: BLE_COMMANDS,
    Domain.Esb: ESB_COMMANDS,
    Domain.Dot15d4: DOT15D4_COMMANDS,
    Domain.LogitechUnifying: UNIFYING_COMMANDS,
    Domain.Phy: PHY_COMMANDS

}


def get_readable_capabilities(caps):
    capabilities = []
    for i in range(24):
        if caps & (1 << i):
            capabilities.append(CAPABILITIES[caps & (1 << i)])
    return capabilities

def get_domain_supported_commands(domain, commands):
    supp_commands = []
    if domain in COMMANDS:
        for i in COMMANDS[domain].keys():
            if commands & (1 << i):
                supp_commands.append(COMMANDS[domain][i])
    return supp_commands

def info(message):
    print_formatted_text(HTML('<ansicyan>[i]</ansicyan> ' + escape(message)))

def error(message):
    print_formatted_text(HTML('<ansired>[e]</ansired> ' + escape(message)))

def main():
    if len(sys.argv) >= 2:
        # Retrieve target interface
        interface = sys.argv[1]

        # Connect to target device and performs discovery

        try:
            dev = WhadDevice.create(interface)

            info('Connecting to device ...')
            dev.open()
            dev.discover()

            info('Device details')
            print('')
            print_formatted_text(
                HTML('<b> Device ID:</b> %s' % ":".join(["{:02x}".format(i) for i in dev.device_id.encode()]))
            )
            print_formatted_text(
                HTML('<b> Firmware info:</b> ')
            )
            if len(dev.info.fw_author) > 0:
                print_formatted_text(
                    HTML('<b> - Author :</b> %s' % escape(dev.info.fw_author))
                )
            if len(dev.info.fw_url) > 0:
                print_formatted_text(
                    HTML('<b> - URL :</b> %s' % escape(dev.info.fw_url))
                )
            print_formatted_text(
                HTML('<b> - Version :</b> %s' % escape(dev.info.version_str))
            )
            print('')
            try:
                info('Discovering domains ...')
                domains = {}
                for domain in dev.get_domains():
                    if domain in DOMAINS:
                        caps_val = dev.get_domain_capability(domain)
                        domains[domain] = get_readable_capabilities(caps_val)
                info('Domains discovered.')
                print('')

                for domain in domains:
                    print_formatted_text(HTML(' <b>This device supports %s:</b>' % DOMAINS[domain]))
                    for cap in domains[domain]:
                        print(' - %s' % cap)
                    print('')
                    print_formatted_text(HTML(' <b>List of supported commands:</b>'))
                    for cmd in get_domain_supported_commands(domain, dev.get_domain_commands(domain)):
                        try:
                            command_name, command_desc = cmd.split(":")
                            print_formatted_text(
                                HTML('  - <b>%s</b>: %s' % (
                                    command_name,
                                    command_desc[1:]
                                    )
                                )
                            )
                        except:
                            pass

                    print('')
            except Exception as err:
                error('[e] An error occured while requesting this device.' +
                      'We were not able to retrieve the supported domains.')

            dev.close()
        except WhadDeviceNotFound:
            error('Device not found')
            exit(1)
        except WhadDeviceNotReady:
            error('Cannot communicate with the device. Make sure it is a ' +
                  'WHAD compatible device and reset it.')
        except PermissionError:
            error('Cannot access the requested device (permission error).')
    else:
        info("Available devices")
        for device in WhadDevice.list(): #print('Usage: %s [device]' % sys.argv[0])
            print_formatted_text(HTML("- <b>%s</b>" % device.interface))
            print_formatted_text(HTML("  <b>Type</b>: %s" % device.type))
            print_formatted_text(HTML("  <b>Index</b>: %s" % device.index))
            print_formatted_text(HTML("  <b>Identifier</b>: %s" % device.identifier))
            print()

if __name__ == '__main__':
    main()



"""BLE Scan command handler
"""

from prompt_toolkit import print_formatted_text, HTML
from whad.cli.app import command
from whad.ble import Scanner

@command('scan')
def scan_handler(app, command_args):
    """scan for BLE devices

    <ansicyan><b>scan</b></ansicyan>

    This command will scan for BLE devices and show them in a list.
    """
    # We need to have an interface specified
    if app.is_piped_interface():
        app.error('This command cannot be used chained with another whad tool.')        
    elif app.interface is not None:
        # Switch to BLE scan mode
        scanner = Scanner(app.interface)
    
        print_formatted_text(HTML('<ansigreen> RSSI Lvl  Type  BD Address        Extra info</ansigreen>'))
        scanner.start()
        try:
            for device in scanner.discover_devices():
                # Show device
                print(device)
        except KeyboardInterrupt as keybd_int:
            print('\rScan terminated by user')
        scanner.stop()



"""WHAD server tool

This utility implements a server module, allowing to create a TCP proxy
which can be used to access a device remotely.
"""
import logging
from prompt_toolkit import print_formatted_text, HTML
import time
from whad.common.monitors.pcap import PcapWriterMonitor
from whad.cli.app import CommandLinePipe, CommandLineApp, ApplicationError, run_app
from scapy.all import *
from whad.common.pcap import extract_pcap_metadata
import sys, os, stat
from whad.device import WhadDevice
from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady
from whad.tools.utils import list_implemented_sniffers, build_configuration_from_args
from whad.tools.wsniff import WhadSniffApp, WhadDomainSubParser

logger = logging.getLogger(__name__)

class WhadPlayApp(WhadSniffApp):

    def __init__(self):
        """Application uses an interface and has commands.
        """
        super().__init__(
            description='WHAD play tool',
            interface=False,
            pcap_argument=True
        )

        self.add_argument(
            "--flush",
            dest="flush",
            action="store_true",

        )

    def infer_domain_from_pcap(self):
        self.pcap_file = None
        index_pcap_file = None
        override_domain = False
        for i in range(len(sys.argv)):
            if ".pcap" in sys.argv[i]:
                self.pcap_file = sys.argv[i]
                index_pcap_file = i
            elif sys.argv[i] in list_implemented_sniffers().keys():
                override_domain = True
        if index_pcap_file is not None and not override_domain:
            if os.path.exists(self.pcap_file):
                domain = extract_pcap_metadata(self.pcap_file)
                if domain in list_implemented_sniffers().keys():
                    sys.argv.insert(index_pcap_file + 1, domain)
                else:
                    self.error("You need to provide a domain")
                    exit(1)
            else:
                self.error("PCAP file not found")
                exit(1)

    def pre_run(self):
        """Pre-run operations: configure scapy theme.
        """

        # If no color is not selected, configure scapy color theme
        self.infer_domain_from_pcap()
        super().pre_run()

        if not self.args.nocolor:
            conf.color_theme = BrightTheme()

        if self.args.pcap is not None:
            self.interface = WhadDevice.create("pcap:" + ("flush:" if self.args.flush else "") + self.args.pcap )

    def run(self):
        try:
            super().run()
        except WhadDeviceNotFound:
            self.error("PCAP file not found")

from whad.exceptions import WhadDeviceAccessDenied, WhadDeviceNotFound
def wplay_main():
    app = WhadPlayApp()
    run_app(app)



from whad.common.analyzer import TrafficAnalyzer
from whad.unifying.stack.constants import ClickType
from whad.unifying.hid import LogitechUnifyingMouseMovementConverter
from whad.scapy.layers.unifying import Logitech_Mouse_Payload, Logitech_Unencrypted_Keystroke_Payload, Logitech_Encrypted_Keystroke_Payload
from whad.unifying.hid import LogitechUnifyingKeystrokeConverter, HIDCodeNotFound, InvalidHIDData
from whad.unifying.crypto import LogitechUnifyingKeyDerivation

class UnifyingMouseMovement(TrafficAnalyzer):
        def reset(self):
            super().reset()
            self.x = None
            self.y = None
            self.wheel_x = None
            self.wheel_y = None
            self.button = None

        @property
        def output(self):
            return {
                "x" : self.x,
                "y" : self.y,
                "wheel_x" : self.wheel_x,
                "wheel_y" : self.wheel_y,
                "button" : self.button
            }

        def process_packet(self, packet):
            if Logitech_Mouse_Payload in packet:
                self.trigger()
                self.mark_packet(packet)
                converter = LogitechUnifyingMouseMovementConverter()
                self.x, self.y = converter.get_coordinates_from_hid_data(packet.movement)
                self.wheel_x, self.wheel_y = packet.wheel_x, packet.wheel_y
                self.button = ClickType(packet.button_mask)
                self.complete()


class UnifyingKeystroke(TrafficAnalyzer):

        @property
        def output(self):
            return {
                "key" :self.key
            }

        def process_packet(self, packet):
            hid_data = None
            if Logitech_Unencrypted_Keystroke_Payload in packet:
                self.trigger()
                self.mark_packet(packet)
                hid_data = packet.hid_data

            if Logitech_Encrypted_Keystroke_Payload in packet:
                self.trigger()
                self.mark_packet(packet)
                if hasattr(packet, "decrypted") and packet.decrypted is not None:
                    hid_data = packet.decrypted.hid_data
                else:
                    hid_data = packet.hid_data

            if hid_data is not None:
                if hid_data == b"\x00" * 7:
                    self.key = None
                else:
                    try:
                        key = LogitechUnifyingKeystrokeConverter.get_key_from_hid_data(hid_data, locale=self.__locale)
                        if key != self.key:
                            self.key = key
                            if len(self.key) > 1:
                                self.key = " [{}] ".format(self.key)
                            self.complete()
                    except (HIDCodeNotFound, InvalidHIDData):
                        pass

        def reset(self):
            super().reset()
            self.__locale = "fr"
            self.key = None


analyzers = {
    "pairing_cracking" : LogitechUnifyingKeyDerivation,
    "mouse" : UnifyingMouseMovement,
    "keystroke" : UnifyingKeystroke
}



"""
Bluetooth Low Energy
"""
import re
from time import time
from whad.ble.stack.gatt import GattClient, GattServer
from whad.helpers import message_filter
from whad.common.triggers import ManualTrigger, ConnectionEventTrigger, ReceptionTrigger
from whad.exceptions import UnsupportedDomain, UnsupportedCapability

from whad.protocol.ble.ble_pb2 import BleDirection, BleAdvType, Connected
from whad.ble.stack import BleStack
from scapy.layers.bluetooth4LE import BTLE_ADV_IND, \
    BTLE_ADV_NONCONN_IND, BTLE_ADV_DIRECT_IND, BTLE_ADV_SCAN_IND, BTLE_SCAN_RSP,\
    BTLE_ADV
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.profile.device import PeripheralDevice
from whad.ble.profile import GenericProfile
from whad.ble.profile.advdata import AdvDataFieldList, AdvFlagsField, AdvDataField, AdvCompleteLocalName, \
    AdvManufacturerSpecificData, AdvShortenedLocalName, AdvTxPowerLevel, AdvDataFieldListOverflow, AdvDataError
from whad.ble.connector import BLE, Central, Peripheral, Sniffer, Hijacker, Injector, Scanner, PeripheralClient
from whad.ble.utils.phy import PHYS
from whad.scapy.layers.bt_mesh import BTMesh_Unprovisioned_Device_Beacon, EIR_BTMesh_Beacon

def is_bdaddr_valid(bd_addr):
    return re.match('^([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$',bd_addr)

__all__ = [
    'GattClient',
    'GattServer',
    'BDAddress',
    'GenericProfile',
    'AdvDataFieldList',
    'AdvFlagsField',
    'AdvDataField',
    'AdvCompleteLocalName',
    'AdvManufacturerSpecificData',
    'AdvShortenedLocalName',
    'AdvTxPowerLevel',
    'AdvDataFieldListOverflow',
    'AdvDataError',
    'Central',
    'Peripheral',
    'PeripheralClient',
    'Sniffer',
    'Hijacker',
    'Injector',
    'Scanner',
    'PHYS',
    'ConnectionEventTrigger',
    'ManualTrigger',
    'ReceptionTrigger'
]



from whad.exceptions import WhadDeviceNotFound, WhadDeviceNotReady, WhadDeviceAccessDenied, \
    WhadDeviceUnsupportedOperation
from whad.device.virtual import VirtualDevice
from whad.hub.discovery import Domain

from whad.hub.generic.cmdresult import CommandResult
from whad.hub.discovery import Capability
from whad.hub.ble import Direction as BleDirection, Commands, AddressType

from whad.scapy.layers.bluetooth import BluetoothUserSocketFixed
from scapy.layers.bluetooth import BluetoothSocketError, \
    HCI_Hdr, HCI_Command_Hdr, HCI_Cmd_Reset, HCI_Cmd_Set_Event_Filter, \
    HCI_Cmd_Connect_Accept_Timeout, HCI_Cmd_Set_Event_Mask, HCI_Cmd_LE_Host_Supported, \
    HCI_Cmd_Read_BD_Addr, HCI_Cmd_Complete_Read_BD_Addr, HCI_Cmd_LE_Set_Scan_Enable, \
    HCI_Cmd_LE_Set_Scan_Parameters, HCI_Cmd_LE_Create_Connection, HCI_Cmd_Disconnect, \
    HCI_Cmd_LE_Set_Advertise_Enable, HCI_Cmd_LE_Set_Advertising_Data, HCI_Event_Disconnection_Complete, \
    HCI_Cmd_LE_Set_Scan_Response_Data, HCI_Cmd_LE_Set_Random_Address, HCI_Cmd_LE_Long_Term_Key_Request_Reply,\
    HCI_Cmd_LE_Start_Encryption_Request

from whad.device.virtual.hci.converter import HCIConverter
from whad.device.virtual.hci.hciconfig import HCIConfig
from whad.device.virtual.hci.constants import LE_STATES, ADDRESS_MODIFICATION_VENDORS, HCIInternalState
from whad.scapy.layers.hci import HCI_Cmd_Read_Local_Version_Information, \
    HCI_Cmd_Complete_Read_Local_Version_Information, HCI_VERSIONS, BT_MANUFACTURERS, \
    HCI_Cmd_Read_Local_Name, HCI_Cmd_Complete_Read_Local_Name, HCI_Cmd_LE_Read_Supported_States, \
    HCI_Cmd_Complete_LE_Read_Supported_States, HCI_Cmd_CSR_Write_BD_Address, HCI_Cmd_CSR_Reset, \
    HCI_Cmd_TI_Write_BD_Address, HCI_Cmd_BCM_Write_BD_Address, HCI_Cmd_Zeevo_Write_BD_Address, \
    HCI_Cmd_Ericsson_Write_BD_Address, HCI_Cmd_ST_Write_BD_Address, HCI_Cmd_LE_Set_Host_Channel_Classification
from select import select
from os import read, write
from time import sleep
from queue import Queue, Empty
from struct import unpack

import logging
logger = logging.getLogger(__name__)

def get_hci(index):
    '''
    Returns an HCI socket based on adapter index.
    '''
    try:
        logger.debug('Creating bluetooth socket ...')
        socket = BluetoothUserSocketFixed(index)
        logger.debug('Bluetooth socket successfully created.')
        return socket
    except BluetoothSocketError:
        logger.debug('An error occured while creating bluetooth socket')
        try:
            logger.debug("Shutting down HCI interface #%d", index)
            HCIConfig.down(index)
            logger.debug("HCI interface %d shut down, creating Bluetooth socket ...", index)
            socket = BluetoothUserSocketFixed(index)
            logger.debug("Bluetooth socket successfully created.")
            return socket
        except BluetoothSocketError as err:
            logger.debug(err)
            logger.debug("Cannot create Bluetooth socket !")
            return None
        except PermissionError as perm_err:
            logger.debug("WHAD device hci%d cannot be accessed.", index)
            raise WhadDeviceAccessDenied("hci%d" % index) from perm_err
    except PermissionError:
        logger.debug("WHAD device hci%d cannot be accessed.", index)
        raise WhadDeviceAccessDenied("hci%d" % index)


class HCIDevice(VirtualDevice):

    INTERFACE_NAME = "hci"

    @classmethod
    def list(cls):
        '''
        Returns a list of available HCI devices.
        '''
        available_devices = []
        devices_ids = HCIConfig.list()
        for device_id in devices_ids:
            available_devices.append(HCIDevice(index=device_id))

        return available_devices

    def __init__(self, index):
        super().__init__()
        self.__converter = HCIConverter(self)
        self.__index = index
        self._advertising = False
        self.__socket = None
        self.__internal_state = HCIInternalState.NONE
        self.__opened = False
        self.__hci_responses = Queue()
        self._dev_capabilities = None
        self._bd_address = None
        self._bd_address_type = AddressType.PUBLIC
        self._fw_version = None
        self._fw_url = None
        self._fw_author = None
        self._dev_id = None
        self._manufacturer = None
        self._cached_scan_data = None
        self._cached_scan_response_data = None
        self.__timeout = 1.0
        self._connected = False
        self._active_handles = []
        self._waiting_disconnect = False

    @property
    def identifier(self):
        '''
        Returns the identifier of the device (e.g., index).
        '''
        return "hci" + str(self.__index)


    def open(self):
        """
        Open device.
        """
        if not self.__opened:
            self.__socket = get_hci(self.__index)
            if self.__socket is None:
                logger.error('Whad device is not ready.')
                raise WhadDeviceNotReady()
            self.__opened = True

            # Ask parent class to run a background I/O thread
            super().open()
            if not self._initialize():
                raise WhadDeviceNotReady()

    def close(self):
        """
        Close current device.
        """
        # Disconnect if necessary
        for handle in self._active_handles:
            self._disconnect(handle)
        # Ask parent class to stop I/O thread
        logger.debug('Stopping background IO threads ...')
        super().close()

        # Close underlying device.
        if self.__socket is not None:
            logger.debug('Closing Bluetooth socket ...')
            self.__socket.close()
            del self.__socket
            self.__socket = None
        self.__opened = False


    def write(self, data):
        """
        Writes data to the device. It relies on select() in order to make sure
        we are allowed to write to the device and wait without eating too much CPU
        if the device is not ready to be written to.

        :param bytes data: Data to write
        :returns: number of bytes written to the device
        """
        if not self.__opened:
            raise WhadDeviceNotReady()
        self.__socket.send(data)

    def read(self):
        """
        Fetches data from the device, if there is any data to read. We call select()
        to make sure data is waiting to be read before reading it.
        """

        if not self.__opened:
            raise WhadDeviceNotReady()
        try:
            if self.__socket is not None and self.__socket.readable(0.1):
                event = self.__socket.recv()
                if event.type == 0x4 and (event.code == 0xe or event.code == 0xf or event.code == 0x13):
                    self.__hci_responses.put(event)
                else:
                    messages = self.__converter.process_event(event)
                    if messages is not None:
                        for message in messages:
                            self._send_whad_message(message)
                    # If the connection is stopped and peripheral mode is started,
                    # automatically re-enable advertising based on cached data
                    if HCI_Event_Disconnection_Complete in event:
                        self._waiting_disconnect = False
                    if HCI_Event_Disconnection_Complete in event and self.__internal_state == HCIInternalState.PERIPHERAL:
                        # If advertising was not enabled, skip
                        if not self._advertising:
                            return

                        # if data are cached, configure them
                        if self._cached_scan_data is not None:
                            # We can't wait for response because we are in the reception loop context
                            success = self._set_advertising_data(self._cached_scan_data, wait_response=False)

                        if self._cached_scan_response_data is not None:
                            success = self._set_scan_response_data(self._cached_scan_response_data, wait_response=False)

                        # We need to artificially disable advertising indicator to prevent cached operation
                        self._advertising = False
                        self._set_advertising_mode(True, wait_response=False)


        except (BrokenPipeError, OSError) as err:
            print(err)
            logger.error("Error, waiting...")
            sleep(1)

    def _wait_response(self, timeout=None):
        response = None
        try:
            response = self.__hci_responses.get(block=True, timeout=timeout)
        except Empty as err:
            logger.debug('[hci] device did not respond')
        return response

    def _write_packet(self, packet):
        """
        Writes an HCI packet.
        """
        logger.debug('[hci] sending packet ...')
        self.__socket.send(packet)

        # Wait for response
        logger.debug('[hci] waiting for response (timeout: %s)...' % self.__timeout)
        response = self._wait_response(timeout=self.__timeout)
        if response is None:
            return False
        else:
            logger.debug('[hci] response code: 0x%04x' % response.code)
            while response.code != 0x13:
                logger.debug('[hci] waiting for response ...')
                response = self._wait_response(timeout=self.__timeout)
                if response is None:
                    return False
                logger.debug('[hci] response code: 0x%04x' % response.code)
        return response.number == 1

    def _write_command(self, command, wait_response=True):
        """
        Writes an HCI command and returns the response.
        """
        hci_command = HCI_Hdr()/HCI_Command_Hdr()/command
        self.__socket.send(hci_command)
        if wait_response:
            response = self._wait_response()
            while response.opcode != hci_command.opcode:
                response = self._wait_response()
        else:
            response = None
        return response

    def reset(self):
        self._bd_address = self._read_BD_address()
        self._local_name = self._read_local_name()
        self._fw_version, self._manufacturer = self._read_local_version_information()
        self._fw_author = self._manufacturer
        self._dev_id = self._generate_dev_id()
        self._fw_url = b"<unknown>"
        self._dev_capabilities = self._get_capabilities()

    def _generate_dev_id(self):
        devid = (bytes.fromhex(self._bd_address.replace(":","")) + self._local_name)[:16]
        if len(devid) < 16:
            devid += b"\x00" * (16 - len(devid))
        return devid

    def _get_capabilities(self):
        supported_states = self._read_LE_supported_states()

        capabilities = 0
        supported_commands = []
        for state in supported_states:
            name, cap, commands = state
            capabilities = capabilities | cap
            supported_commands += commands

        supported_commands += [Commands.SetBdAddress]

        supported_commands = list(set(supported_commands))
        capabilities = {
            Domain.BtLE : (
                                capabilities | Capability.NoRawData,
                                supported_commands
            )
        }
        return capabilities

    def _reset(self):
        """
        Reset HCI device.
        """
        response = self._write_command(HCI_Cmd_Reset())
        return response is not None and response.status == 0x0


    def _set_event_filter(self, type=0):
        """
        Configure HCI device event filter.
        """
        response = self._write_command(HCI_Cmd_Set_Event_Filter(type=type))
        return response is not None and response.status == 0x00

    def _set_event_mask(self, mask=b"\xff\xff\xfb\xff\x07\xf8\xbf\x3d"):
        """
        Configure HCI device event mask.
        """
        response = self._write_command(HCI_Cmd_Set_Event_Mask(mask=mask))
        return response is not None and response.status == 0x00

    def _set_connection_accept_timeout(self, timeout=32000):
        """
        Configure HCI device connection accept timeout.
        """
        response = self._write_command(HCI_Cmd_Connect_Accept_Timeout(timeout=timeout))
        return response is not None and response.status == 0x00

    def _indicates_LE_support(self):
        """
        Indicates to HCI Device that the Host supports Low Energy mode.
        """
        response = self._write_command(HCI_Cmd_LE_Host_Supported())
        return response is not None and response.status == 0x00

    def _initialize(self):
        """
        Initialize HCI Device and returns boolean indicating if it can be used by WHAD.
        """
        success = (
                self._reset() and
                self._set_event_filter(0) and
                self._set_connection_accept_timeout(32000) and
                self._set_event_mask(b"\xff\xff\xfb\xff\x07\xf8\xbf\x3d") and
                self._indicates_LE_support()
        )

        return success

    def _read_BD_address(self):
        """
        Read BD Address used by the HCI device.
        """
        response = self._write_command(HCI_Cmd_Read_BD_Addr())
        if response.status == 0x00 and HCI_Cmd_Complete_Read_BD_Addr in response:
            return response.addr
        return None

    def _read_local_name(self):
        """
        Read local name used by the HCI device.
        """
        response = self._write_command(HCI_Cmd_Read_Local_Name())
        if response.status == 0x00 and HCI_Cmd_Complete_Read_Local_Name in response:
            return response.local_name
        return None

    def _read_local_version_information(self):
        """
        Read local version information used by the HCI device.
        """
        response = self._write_command(HCI_Cmd_Read_Local_Version_Information())
        if response.status == 0x00 and HCI_Cmd_Complete_Read_Local_Version_Information in response:
            version = [int(v) for v in HCI_VERSIONS[response.hci_version].split(".")]
            version += [response.hci_subversion]
            manufacturer = BT_MANUFACTURERS[response.company_identifier].encode("utf-8")
            return version, manufacturer
        return None

    def _read_LE_supported_states(self):
        """
        Returns the list of Bluetooth Low Energy states supported by the HCI device.
        """
        response = self._write_command(HCI_Cmd_LE_Read_Supported_States())
        if response.status == 0x00 and HCI_Cmd_Complete_LE_Read_Supported_States in response:
            states = []
            for bit_position, state in LE_STATES.items():
                if response.supported_states & (1 << bit_position) != 0:
                    states.append(state)
            return states
        else:
            return None

    def _set_BD_address(self, bd_address="11:22:33:44:55:66", bd_address_type=AddressType.PUBLIC):
        """
        Modify the BD address (if supported by the HCI device).
        """
        if bd_address_type == AddressType.PUBLIC:
            _, self._manufacturer = self._read_local_version_information()
            if self._manufacturer in ADDRESS_MODIFICATION_VENDORS:
                logger.info("[i] Address modification supported !")
                if self._manufacturer == b'Qualcomm Technologies International, Ltd. (QTIL)':
                    # Keep in cache existing devices
                    existing_devices = devices = HCIConfig.list()

                    # Write BD address and reset with vendor specific commands
                    self._write_command(HCI_Cmd_CSR_Write_BD_Address(addr=bd_address), wait_response=False)
                    self._write_command(HCI_Cmd_CSR_Reset(), wait_response=False)

                    # We are forced to close the socket and reopen it here...
                    self.__socket.close()
                    # Add a delay to prevent error
                    sleep(0.5)

                    # The index may have changed, find it automatically and reconfigure self.__index
                    success = False
                    while not success:
                        devices = HCIConfig.list()
                        if self.__index not in devices:
                            for i in existing_devices:
                                if i != self.__index:
                                    devices.remove(i)
                            if len(devices) > 0:
                                self.__index = devices[0]
                                success = True

                    # If all goes right, we should be able to open a new socket
                    self.__socket = get_hci(self.__index)
                    # Initialize a new socket
                    self._initialize()

                else:
                    # For the other manufacturers, we only need to pick the right command and perform a reset
                    BD_ADDRESS_MODIFICATION_MAP = {
                        b'Texas Instruments Inc.' : HCI_Cmd_TI_Write_BD_Address,
                        b'Broadcom Corporation' : HCI_Cmd_BCM_Write_BD_Address,
                        b'Zeevo, Inc.' : HCI_Cmd_Zeevo_Write_BD_Address,
                        b'Ericsson Technology Licensing' : HCI_Cmd_Ericsson_Write_BD_Address,
                        b'Integrated System Solution Corp.' : HCI_Cmd_Ericsson_Write_BD_Address,
                        b'ST Microelectronics' : HCI_Cmd_ST_Write_BD_Address
                    }
                    command = BD_ADDRESS_MODIFICATION_MAP[self._manufacturer]
                    self._write_command(command, wait_response=False)
                    self._reset()

                # Check the modification success and re-generate device ID
                self._bd_address = self._read_BD_address()
                self._dev_id = self._generate_dev_id()
                self._bd_address_type = AddressType.PUBLIC
                return self._bd_address == bd_address

            else:
                logger.debug("Address modification not supported.")
                return False
        else:
            self._write_command(HCI_Cmd_LE_Set_Random_Address(address=bd_address))
            self._bd_address = self._read_BD_address()
            self._bd_address_type = AddressType.RANDOM
            return True

    def _set_scan_parameters(self, active=True):
        """
        Configure Scan parameters for HCI device.
        """
        response = self._write_command(HCI_Cmd_LE_Set_Scan_Parameters(type=int(active)))
        return response is not None and response.status == 0x00

    def _set_scan_mode(self, enable=True):
        """
        Enable or disable scan mode for HCI device.
        """
        response = self._write_command(HCI_Cmd_LE_Set_Scan_Enable(enable=int(enable), filter_dups=False))
        return response is not None and response.status == 0x00

    def _connect(self, bd_address, bd_address_type=AddressType.PUBLIC, hop_interval=96, channel_map=None):
        """
        Establish a connection using HCI device.
        """
        patype = 0 if bd_address_type == AddressType.PUBLIC else 1
        if channel_map is not None:
            formatted_channel_map = unpack("<Q",channel_map+ b"\x00\x00\x00")[0]
            response = self._write_command(HCI_Cmd_LE_Set_Host_Channel_Classification(chM=formatted_channel_map))
            if response.status != 0x00:
                return False
        response = self._write_command(
            HCI_Cmd_LE_Create_Connection(
                paddr=bd_address,
                patype=patype,
                min_interval=hop_interval,
                max_interval=hop_interval
            )
        )
        return response is not None and response.status == 0x00

    def _disconnect(self, handle):
        """
        Establish a disconnection using HCI device.
        """
        response = self._write_command(HCI_Cmd_Disconnect(handle=handle))
        self._waiting_disconnect = True
        while self._waiting_disconnect:
            sleep(0.1)
        return response is not None and response.status == 0x00

    def _set_advertising_data(self, data, wait_response=True):
        """
        Configure advertising data to use by HCI device.
        """
        # pad data if less than 31 bytes
        if len(data) < 31:
            data += b'\x00'*(31 - len(data))

        # Send command
        if wait_response:
            response = self._write_command(HCI_Cmd_LE_Set_Advertising_Data(data=data))
            return response is not None and response.status == 0x0
        else:
            self._write_command(HCI_Cmd_LE_Set_Advertising_Data(data=data), wait_response=False)
            return True

    def _set_scan_response_data(self, data, wait_response=True):
        """
        Configure scan response data to use by HCI device.
        """
        if wait_response:
            response = self._write_command(HCI_Cmd_LE_Set_Scan_Response_Data(
                    data=data + (31 - len(data)) * b"\x00", len=len(data)
                )
            )
            return response is not None and response.status == 0x0
        else:
            response = self._write_command(HCI_Cmd_LE_Set_Scan_Response_Data(
                    data=data + (31 - len(data)) * b"\x00", len=len(data)
                ),
                wait_response=False
            )
            return True


    def _set_advertising_mode(self, enable=True, wait_response=True):
        """
        Enable or disable advertising mode for HCI device.
        """
        if self._advertising and enable:
            return True
        else:
            logger.debug('Enable advertising: %s' % enable)
            if wait_response:
                response = self._write_command(HCI_Cmd_LE_Set_Advertise_Enable(enable=int(enable)))
                success = response.status == 0x00
            else:
                self._write_command(HCI_Cmd_LE_Set_Advertise_Enable(enable=int(enable)), wait_response=False)
                success = True
            if success:
                self._advertising = enable
            return success

    def _enable_encryption(self, enable=True, handle=None,  key=None, rand=None, ediv=None):

        if self.__converter.pending_key_request:
            response = self._write_command(
                HCI_Cmd_LE_Long_Term_Key_Request_Reply(
                    handle=handle,
                    ltk=key[::-1]
                )
            )
            self.__converter.pending_key_request = False
        else:
            response = self._write_command(
                HCI_Cmd_LE_Start_Encryption_Request(
                    handle=handle,
                    ltk=key[::-1],
                    rand=rand,
                    ediv=unpack('<H', ediv)[0]
                )
            )
        return response.status == 0x00


    def _on_whad_ble_encryption(self, message):
        success = self._enable_encryption(
            message.enabled,
            message.conn_handle,
            message.key,
            message.rand,
            message.ediv
        )
        if success:
            self._send_whad_command_result(CommandResult.SUCCESS)
            return
        self._send_whad_command_result(CommandResult.ERROR)


    def _on_whad_ble_periph_mode(self, message):
        logger.debug('whad ble periph mode message')
        if Commands.PeripheralMode in self._dev_capabilities[Domain.BtLE][1]:
            success = True
            if len(message.scan_data) > 0:
                success = success and self._set_advertising_data(message.scan_data)
                self._cached_scan_data = message.scan_data
            if len(message.scanrsp_data) > 0:
                success = success and self._set_scan_response_data(message.scanrsp_data)
                self._cached_scan_response_data = message.scanrsp_data
            success = success and self._set_advertising_mode(True)
            if success:
                self.__internal_state = HCIInternalState.PERIPHERAL
                self._send_whad_command_result(CommandResult.SUCCESS)
                return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_disconnect(self, message):
        success = self._disconnect(message.conn_handle)
        if success:
            self._send_whad_command_result(CommandResult.SUCCESS)
            return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_connect(self, message):
        if Commands.ConnectTo in self._dev_capabilities[Domain.BtLE][1]:
            bd_address = message.bd_address
            bd_address_type = message.addr_type
            channel_map = message.channel_map if message.channel_map is not None else None
            hop_interval = message.hop_interval if message.hop_interval is not None else 96
            if self._connect(bd_address, bd_address_type, hop_interval=hop_interval, channel_map=channel_map):
                self._send_whad_command_result(CommandResult.SUCCESS)
                return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_central_mode(self, message):
        if Commands.CentralMode in self._dev_capabilities[Domain.BtLE][1]:
            self.__internal_state = HCIInternalState.CENTRAL
            self._send_whad_command_result(CommandResult.SUCCESS)
            return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_scan_mode(self, message):
        if Commands.ScanMode in self._dev_capabilities[Domain.BtLE][1]:
            active_scan = message.active
            if self._set_scan_parameters(active_scan):
                self.__internal_state = HCIInternalState.SCANNING
                self._send_whad_command_result(CommandResult.SUCCESS)
                return
        self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_start(self, message):
        logger.info('whad internal state: %d' % self.__internal_state)
        if self.__internal_state == HCIInternalState.SCANNING:
            self._set_scan_mode(True)
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state == HCIInternalState.PERIPHERAL:
            if not self._advertising:
                if self._set_advertising_mode(True):
                    self._advertising = True
                    self._send_whad_command_result(CommandResult.SUCCESS)
                else:
                    self._send_whad_command_result(CommandResult.ERROR)
            else:
                self._send_whad_command_result(CommandResult.SUCCESS)

        elif self.__internal_state == HCIInternalState.CENTRAL:
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_stop(self, message):
        if self.__internal_state == HCIInternalState.SCANNING:
            self._set_scan_mode(False)
            self.__internal_state = HCIInternalState.NONE
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state == HCIInternalState.CENTRAL:
            self.__internal_state = HCIInternalState.NONE
            self._send_whad_command_result(CommandResult.SUCCESS)
        elif self.__internal_state == HCIInternalState.PERIPHERAL:
            # We are not advertising anymore
            self._advertising = False
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_send_pdu(self, message):
        logger.debug('Received WHAD BLE send_pdu message')
        if ((self.__internal_state == HCIInternalState.CENTRAL and message.direction == BleDirection.MASTER_TO_SLAVE) or
           (self.__internal_state == HCIInternalState.PERIPHERAL and message.direction == BleDirection.SLAVE_TO_MASTER)):
            try:
                hci_packets = self.__converter.process_message(message)

                if hci_packets is not None:
                    logger.debug('sending HCI packets ...')
                    success = True
                    for hci_packet in hci_packets:
                        success = success and self._write_packet(hci_packet)
                    logger.debug('HCI packet sending result: %s' % success)
                    if success:
                        logger.debug('send_pdu command succeeded.')
                        self._send_whad_command_result(CommandResult.SUCCESS)
                    else:
                        logger.debug('send_pdu command failed.')
                        self._send_whad_command_result(CommandResult.ERROR)
                pending_messages = self.__converter.get_pending_messages()
                for pending_message in pending_messages:
                    self._send_whad_message(pending_message)

            except WhadDeviceUnsupportedOperation as err:
                logger.debug('Parameter error')
                self._send_whad_command_result(CommandResult.PARAMETER_ERROR)
        else:
            # Wrong state, cannot send PDU
            logger.debug('wrong state or packet direction.')
            self._send_whad_command_result(CommandResult.ERROR)

    def _on_whad_ble_set_bd_addr(self, message):
        logger.debug('Received WHAD BLE set_bd_addr message')
        if self._set_BD_address(message.bd_address, message.addr_type):
            logger.debug('HCI adapter BD address set to %s' % str(message.bd_address))
            self._send_whad_command_result(CommandResult.SUCCESS)
        else:
            logger.debug('HCI adapter does not support BD address spoofing')
            self._send_whad_command_result(CommandResult.ERROR)



"""Bluetooth Low Energy Battery Service Profile
"""
from struct import pack, unpack
from whad.ble.profile.attribute import UUID
from whad.ble.profile import PrimaryService, Characteristic

class BatteryService(object):
    """Battery Service Profile

    This service implements the BLE Battery Service as described in the spec.
    """

    battery = PrimaryService(
        uuid = UUID(0x180f),
        level = Characteristic(
            uuid = UUID(0x2A19),
            permissions = ['read'],
            notify = True,
            indicate = True,
            value=pack('B', 100)
        )
    )

    def set_battery_level(self, level: int):
        """Set battery level
        """
        if level >= 0 and level<=100:
            self.battery.level.value = pack('B', level)

    def get_battery_level(self):
        """Return battery level
        """
        return unpack('B', self.battery.level.value)[0]


from whad.rf4ce.connector import RF4CE
from whad.rf4ce.sniffing import SnifferConfiguration
from whad.rf4ce.crypto import RF4CEDecryptor, RF4CEKeyDerivation
from whad.rf4ce.utils.adpcm import ADPCM
from whad.rf4ce.exceptions import MissingCryptographicMaterial
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter, is_message_type
from whad.rf4ce.sniffing import KeyExtractedEvent
from whad.common.sniffing import EventsManager
from whad.hub.dot15d4 import RawPduReceived, PduReceived
from whad.hub.message import AbstractPacket
from whad.exceptions import WhadDeviceDisconnected
import logging

logger = logging.getLogger(__name__)

class Sniffer(RF4CE, EventsManager):
    """
    RF4CE Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device, configuration=SnifferConfiguration()):
        RF4CE.__init__(self, device)
        EventsManager.__init__(self)


        self.__configuration = configuration
        self.__decryptor = RF4CEDecryptor()
        self.__key_derivation = RF4CEKeyDerivation()
        self.__audio_stream = ADPCM()

        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def _enable_sniffing(self):
        for key in self.__configuration.keys:
            self.__decryptor.add_key(key)
        for address in self.__configuration.addresses:
            self.__decryptor.add_address(address)
        if (
            self.__configuration.audio and
            self.__configuration.audio_file is not None and
            self.__configuration.audio_file
        ):
            self.__audio_stream.output_filename = self.__configuration.audio_file
        self.sniff_rf4ce(channel=self.__configuration.channel)

    def add_key(self, key):
        self.__configuration.keys.append(key)

    def clear_keys(self):
        self.__configuration.keys = []

    def add_address(self, key):
        self.__configuration.addresses.append(address)

    def clear_addresses(self):
        self.__configuration.addresses = []

    @property
    def decrypt(self):
        return self.__configuration.decrypt

    @decrypt.setter
    def decrypt(self, decrypt):
        self.__configuration.decrypt = decrypt


    @property
    def channel(self):
        return self.__configuration.channel

    @channel.setter
    def channel(self, channel=15):
        self.stop()
        self.__configuration.channel = channel
        self._enable_sniffing()


    @property
    def configuration(self):
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        #self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    def process_packet(self, packet):

        if self.__configuration.pairing:
            self.__key_derivation.process_packet(packet)
            if self.__key_derivation.key is not None:
                logger.info("[i] New key extracted: ", self.__key_derivation.key.hex())
                self.trigger_event(KeyExtractedEvent(self.__key_derivation.key))
                self.__decryptor.add_key(self.__key_derivation.key)
                self.__key_derivation.reset()

        if (
            hasattr(packet, "fcf_destaddrmode") and
            packet.fcf_destaddrmode == 3
        ):
            self.__decryptor.add_address(packet.dest_addr)

        if (
            hasattr(packet, "fcf_srcaddrmode") and
            packet.fcf_srcaddrmode == 3
        ):
            self.__decryptor.add_address(packet.src_addr)

        if (
            hasattr(packet, "security_enabled") and
            packet.security_enabled == 1 and
            self.__configuration.decrypt
        ):
            try:
                decrypted, success = self.__decryptor.attempt_to_decrypt(packet)
                if success:
                    metadata = packet.metadata
                    packet = decrypted
                    packet.metadata = metadata
                    if self.__configuration.audio:
                        self.__audio_stream.process_packet(packet)

            except MissingCryptographicMaterial:
                pass
        return packet

    def sniff(self):
        try:
            while True:
                if self.support_raw_pdu():
                    message_type = RawPduReceived
                else:
                    message_type = PduReceived

                message = self.wait_for_message(filter=message_filter(message_type), timeout=.1)
                if message is not None and issubclass(message, AbstractPacket):
                    packet = message.to_packet()
                    packet = self.process_packet(packet)

                    self.monitor_packet_rx(packet)

                    yield packet
        except WhadDeviceDisconnected:
            return



"""
Bluetooth LE Stack L2CAP manager
"""
from struct import unpack, pack
from scapy.layers.bluetooth import L2CAP_Hdr, ATT_Hdr, SM_Hdr, L2CAP_CmdHdr, \
    L2CAP_Connection_Parameter_Update_Request, L2CAP_Connection_Parameter_Update_Response
from whad.ble.stack.att import ATTLayer
from whad.ble.stack.smp import SMPLayer

from whad.common.stack import Layer, alias, source, state, LayerState, ContextualLayer

import logging
logger = logging.getLogger(__name__)

@alias('l2cap')
class L2CAPLayer(ContextualLayer):

    def configure(self, options={}):
        # Initialize state
        self.state.conn_handle = None
        self.state.local_mtu = 23
        self.state.remote_mtu = 23
        self.state.fifo = None
        self.state.expected_len = 0

    def set_conn_handle(self, conn_handle):
        '''Save current connection handle
        '''
        self.state.conn_handle = conn_handle

    def get_conn_handle(self):
        """Retrieve current connection handle
        """
        return self.state.conn_handle

    def set_local_mtu(self, mtu):
        '''Set local MTU
        '''
        logger.debug('local MTU changed to %d for conn_handle %d' % (mtu, self.state.conn_handle))
        self.state.local_mtu = mtu

    def set_remote_mtu(self, mtu):
        '''Set remote MTU
        '''
        logger.debug('remote MTU changed to %d for conn_handle %d' % (mtu, self.state.conn_handle))
        self.state.remote_mtu = mtu

    def get_local_mtu(self):
        return self.state.local_mtu

    @source('ll')
    def on_data_received(self, l2cap_data, fragment=False):
        """Handles incoming L2CAP data"""
        if fragment and self.state.fifo is not None:
            logger.debug('[l2cap] Received a L2CAP fragment of %d bytes' % len(l2cap_data))
            self.state.fifo += l2cap_data
            logger.debug('[l2cap] L2CAP packet size so far: %d' % len(self.state.fifo))

            if len(self.state.fifo) >= self.state.expected_len:
                # We have received a complete L2CAP packet, process it
                logger.debug('[l2cap] Received a complete L2CAP packet, process it')
                self.on_l2cap_packet(L2CAP_Hdr(self.state.fifo[:self.state.expected_len]))
                self.state.fifo = None

        elif len(l2cap_data) >= 2:
            # Start of L2CAP or complete L2CAP message
            self.state.fifo = l2cap_data
            logger.debug('[l2cap] received start of fragmented or complete message')

            # Check if we have a complete L2CAP message
            self.state.expected_len = unpack('<H', self.state.fifo[:2])[0] + 4
            logger.debug('[l2cap] expected l2cap length: %d' % self.state.expected_len)
            logger.debug('[l2cap] actual l2cap length: %d' % len(self.state.fifo))
            if len(self.state.fifo) >= self.state.expected_len:
                # We have received a complete L2CAP packet, process it
                logger.debug('[l2cap] Received a complete L2CAP packet, process it')
                self.on_l2cap_packet(L2CAP_Hdr(self.state.fifo[:self.state.expected_len]))
                self.state.fifo = None


    def on_l2cap_packet(self, packet):
        """Process incoming L2CAP packets.
        """
        if ATT_Hdr in packet:
            self.send('att', packet.getlayer(ATT_Hdr))
        elif L2CAP_CmdHdr in packet:
            self.on_cmd_packet(packet.getlayer(L2CAP_CmdHdr))
        elif SM_Hdr in packet:
            self.send('smp', packet.getlayer(SM_Hdr))

    def on_cmd_packet(self, packet):
        """Handle L2CAP Connection Parameter Update Requests.

        This command is rejected by default.
        """
        if L2CAP_Connection_Parameter_Update_Request in packet:
            logger.debug('[l2cap] Received a L2CAP Connection Parameter Update Request, rejecting')
            # Reject it
            self.send('ll', L2CAP_Hdr()/L2CAP_CmdHdr(id=packet[L2CAP_CmdHdr].id)/L2CAP_Connection_Parameter_Update_Response(move_result=1))

    def get_fragments(self, data):
        packets=[]
        # If data is bigger than MTU-1, then split
        if len(data) > self.state.remote_mtu:
            nb_packets = int(len(data)/(self.state.remote_mtu - 1))
            if nb_packets * (self.state.remote_mtu - 1) < len(data):
                nb_packets += 1
            raw_data = bytes(data)
            for i in range(nb_packets):
                packets.append(raw_data[i*(self.state.remote_mtu - 1):(i+1)*(self.state.remote_mtu - 1)])
        else:
            nb_packets = 1
            packets = [
                data
            ]
        return packets

    @source('att')
    def on_att_packet_recv(self, data, channel='attribute'):
        """Process incoming packets from ATT that must be forwarded to link layer.
        """
        packets = self.get_fragments(data)

        # Send packets
        for i, pkt in enumerate(packets):
            # First packet is sent with no fragment flag
            if i == 0:
                # Send packet to link layer
                self.send('ll', L2CAP_Hdr(cid=channel, len=len(data))/pkt, fragment=False)
            else:
                # Send packet to link layer
                self.send('ll', pkt, fragment=True)


    @source('smp')
    def on_smp_packet_recv(self, data, channel=0x06):
        """Process incoming packets from ATT that must be forwarded to link layer.
        """
        # Note: we are forced to hardcode the channel in parameters here, because
        # scapy does not provide a complete dictionary mapping 0x06 to 'smp'.

        packets = self.get_fragments(data)

        # Send packets
        for i, pkt in enumerate(packets):
            # First packet is sent with no fragment flag
            if i == 0:
                # Send packet to link layer
                self.send('ll', L2CAP_Hdr(cid=channel, len=len(data))/pkt, fragment=False)
            else:
                # Send packet to link layer
                self.send('ll', pkt, fragment=True)

L2CAPLayer.add(ATTLayer)
L2CAPLayer.add(SMPLayer)



"""WHAD Protocol Hub

The protocol hub is a set of wrappers for Protobuf messages that provides a way
to support different versions of our communication protocol. This protocol hub
provides a default message parser handling different protocol versions that will
pick the correct message wrapper class to parse it. Message wrappers simplifies
the way protocol buffers messages are created by mapping some of their properties
to protobuf messages fields.
"""
import logging
from typing import Union
from whad.protocol.whad_pb2 import Message
from .registry import Registry

logger = logging.getLogger(__name__)

class ProtocolHub(Registry):
    """WHAD Protocol Hub class

    This class is an interface between all our Python code and the devices, that
    support all the existing versions of the WHAD protocol and handles every
    differences that exist between them in a transparent fashion.
    """

    NAME = 'hub'
    VERSIONS = {}

    def __init__(self, proto_version: int):
        """Instanciate a WHAD protocol hub for a specific version.
        """
        self.__version = proto_version

    @property
    def version(self) -> int:
        return self.__version

    @property
    def generic(self):
        return self.get('generic')

    @property
    def discovery(self):
        return self.get('discovery')

    @property
    def ble(self):
        return self.get('ble')

    @property
    def dot15d4(self):
        return self.get('dot15d4')

    @property
    def phy(self):
        return self.get('phy')

    @property
    def esb(self):
        return self.get('esb')

    @property
    def unifying(self):
        return self.get('unifying')

    def get(self, factory: str):
        return ProtocolHub.bound(factory, self.__version)(self.__version)

    def parse(self, data: Union[Message, bytes]):
        """Parse a serialized WHAD message into an associated object.
        """
        if isinstance(data, bytes):
            # Use protocol buffers to parse our message
            msg = Message()
            msg.ParseFromString(bytes(data))
        elif isinstance(data, Message):
            msg = data
        else:
            return None

        # Only process generic messages
        return ProtocolHub.bound(
            msg.WhichOneof('msg'),
            self.__version).parse(self.__version, msg)

    def convert_packet(self, packet):
        """Convert packet to the corresponding message.
        """
        msg = None
        # We dispatch packets based on their layers
        if self.ble.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is BLE')
            msg = self.ble.convert_packet(packet)
        elif self.dot15d4.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is Dot15d4')
            msg = self.dot15d4.convert_packet(packet)
        elif self.esb.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is ESB')
            msg = self.esb.convert_packet(packet)
        elif self.phy.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is PHY')
            msg = self.phy.convert_packet(packet)
        elif self.unifying.is_packet_compat(packet):
            logger.debug('[hub] convert_packet(): packet is Unifying')
            msg = self.unifying.convert_packet(packet)
        else:
            logger.error('[hub] convert_packet(): packet is unknown !')

        return msg



from .generic import Generic
from .discovery import Discovery
from .ble import BleDomain
from .dot15d4 import Dot15d4Domain
from .phy import PhyDomain
from .esb import EsbDomain
from .unifying import UnifyingDomain



from queue import Queue

from whad.exceptions import WhadDeviceDisconnected
from whad.phy.connector import Phy
from whad.phy import Endianness
from whad.phy.sniffing import SnifferConfiguration
from whad.phy.exceptions import NoModulation
from whad.common.sniffing import EventsManager
from whad.exceptions import UnsupportedCapability
from whad.helpers import message_filter, is_message_type

from whad.hub.phy import PacketReceived, RawPacketReceived
from whad.hub.message import AbstractPacket

# TODO: every sniffer is broken (sniff() method does not catch packets, we
#       have to catch them in on_packet() and put them in a queue)

class Sniffer(Phy, EventsManager):
    """
    Phy Sniffer interface for compatible WHAD device.
    """

    def __init__(self, device):
        Phy.__init__(self, device)
        EventsManager.__init__(self)


        self.__configuration = SnifferConfiguration()

        # Check if device can perform sniffing
        if not self.can_sniff():
            raise UnsupportedCapability("Sniff")

    def _enable_sniffing(self):
        self.set_frequency(self.__configuration.frequency)
        self.set_packet_size(self.__configuration.packet_size)

        # Set data rate for all modulations but LoRa
        if not self.__configuration.lora:
            self.set_datarate(self.__configuration.datarate)

        if self.__configuration.gfsk:
            self.set_gfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.bfsk:
            self.set_bfsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.qfsk:
            self.set_4fsk(deviation=self.__configuration.fsk_configuration.deviation)
        elif self.__configuration.ask:
            self.set_ask()
        elif self.__configuration.bpsk:
            self.set_bpsk()
        elif self.__configuration.qpsk:
            self.set_qpsk()
        elif self.__configuration.lora:
            self.set_lora(
                self.__configuration.lora_configuration.spreading_factor,
                self.__configuration.lora_configuration.coding_rate,
                self.__configuration.lora_configuration.bandwidth,
                12,
                crc=self.__configuration.lora_configuration.enable_crc,
                explicit=self.__configuration.lora_configuration.enable_explicit_mode
            )

        # Set endianness for all modulations but LoRa
        if not self.__configuration.lora:
            self.set_endianness(
                Endianness.LITTLE if
                self.__configuration.little_endian else
                Endianness.BIG
            )

        self.set_sync_word(self.__configuration.sync_word)

        self.sniff_phy()

    @property
    def configuration(self):
        return self.__configuration

    @configuration.setter
    def configuration(self, new_configuration):
        self.stop()
        self.__configuration = new_configuration
        self._enable_sniffing()

    @property
    def frequency(self):
        return self.__configuration.frequency

    @frequency.setter
    def frequency(self, frequency=2402000000):
        self.stop()
        self.__configuration.frequency = frequency
        self._enable_sniffing()

    def available_actions(self, filter=None):
        actions = []
        return [action for action in actions if filter is None or isinstance(action, filter)]

    def sniff(self):
        try:
            while True:
                if self.support_raw_iq_stream():
                    message_type = RawPacketReceived
                else:
                    message_type = PacketReceived
                message = self.wait_for_message(filter=message_filter(message_type), timeout=.1)

                if message is not None and issubclass(message, AbstractPacket):
                    packet = message.to_packet()
                    self.monitor_packet_rx(packet)
                    yield packet
        except WhadDeviceDisconnected:
            return



"""
Bluetooth Low Energy Central connector
======================================
"""

from time import time, sleep
from whad.ble.connector import BLE
from whad.hub.ble import Direction
from whad.hub.ble.bdaddr import BDAddress
from whad.ble.stack import BleStack, BtVersion
from whad.ble.stack.constants import BT_MANUFACTURERS, BT_VERSIONS
from whad.ble.stack.gatt import GattClient
from whad.ble.stack.att import ATTLayer
from whad.ble.stack.smp import CryptographicDatabase
from whad.ble.exceptions import ConnectionLostException, PeripheralNotFound
from whad.ble.profile.device import PeripheralDevice
from whad.common.stack import Layer
from whad.protocol.ble.ble_pb2 import BleDirection
from whad.exceptions import UnsupportedCapability

from binascii import hexlify

import logging
logger = logging.getLogger(__name__)

class Central(BLE):
    """This connector provides a BLE Central role.

    To initiate a connection to a device, just call :meth:`Central.connect` with the target
    BD address and it should return an instance of :class:`whad.ble.profile.device.PeripheralDevice` in return.

    """

    def __init__(self, device, existing_connection = None, from_json=None, stack=BleStack, client=GattClient, security_database=None):
        super().__init__(device)

        """Attach a GATT client if specified in parameter

        If `client` is set to None, the default GATT layer is used, which does
        not provide any client feature at all.
        """
        if client is not None:
            if issubclass(client, Layer) and client.alias == 'gatt':
                ATTLayer.add(client)

        self.__gatt_client = None
        self.__stack = stack(self)
        self.__connected = False
        self.__peripheral = None
        self.__random_addr = False
        self.__profile_json = from_json
        self.__target = None
        self.__local = None
        self.__conn_handle = None

        # If no connection, check if
        if not self.can_be_central():
            raise UnsupportedCapability('Central')

        # Initialize security database
        if security_database is None:
            logger.info('No security database provided to this Peripheral instance, use a default one.')
            self.__security_database = CryptographicDatabase()
        else:
            logger.info('Peripheral will use the provided security database.')
            self.__security_database = security_database

        # If a connection already exists, just feed the stack with the parameters
        if existing_connection is not None:
            self.on_connected(existing_connection)
        else:
            # self.stop() # ButteRFly doesn't support calling stop when spawning central
            self.enable_central_mode()


    @property
    def security_database(self):
        return self.__security_database

    @property
    def local_peer(self) -> BDAddress:
        """Local peer BD address.
        """
        return self.__local

    @property
    def target_peer(self) -> BDAddress:
        """Remote peer BD address.
        """
        return self.__target

    @property
    def conn_handle(self):
        return self.__conn_handle

    @property
    def stack(self):
        '''Return the current stack instance
        '''
        return self.__stack

    def connect(self, bd_address, random=False, timeout=30, access_address=None, channel_map=None, crc_init=None, hop_interval=None, hop_increment=None) -> PeripheralDevice:
        """Connect to a target device

        :param  bd_address:     Bluetooth device address (in format 'xx:xx:xx:xx:xx:xx')
        :type   bd_address:     str
        :param  timeout:        Connection timeout
        :type   timeout:        float
        :param  access_address: Access address to use (optional)
        :type   access_address: int
        :param  channel_map:    Channel map to use (optional)
        :type   channel_map:    int
        :param  crc_init:       CRC Initialization value to use (optional)
        :type   crc_init:       int
        :param  hop_interval:   Hop interval to use (optional)
        :type   hop_interval:   int
        :param  hop_increment:  Hop increment to use (optional)
        :type   hop_increment:  int

        :return: An instance of `PeripheralDevice` on success, `None` on failure.
        :rtype: :class:`whad.ble.profile.device.PeripheralDevice`
        """
        if self.can_connect():
            self.connect_to(
                bd_address,
                random=random,
                access_address=access_address,
                channel_map=channel_map,
                crc_init=crc_init,
                hop_interval=hop_interval,
                hop_increment=hop_increment
            )

            self.start()
            start_time=time()
            while not self.is_connected():
                if time()-start_time >= timeout:
                    raise PeripheralNotFound
                sleep(0.1)
            self.__random_addr = random
            return self.peripheral()
        else:
            # TODO: raise cannot connect
            raise PeripheralNotFound()

    def peripheral(self) -> PeripheralDevice:
        """Connected BLE peripheral.
        """
        return self.__peripheral


    def send_pdu(self, pdu, conn_handle=0, direction=Direction.MASTER_TO_SLAVE, access_address=0x8e89bed6, encrypt=None) -> bool:
        """Send a PDU to the connected peripheral device or to the central device.

        :param  pdu:            BLE PDU to send.
        :type   pdu:            :class:`scapy.layers.bluetooth4LE.BTLE`
        :param  conn_handle:    Connection handle
        :type   conn_handle:    int
        :param  direction:      Direction (central to peripheral, peripheral to central)
        :type   direction:      :class:`whad.protocol.ble.ble_pb2.BleDirection`
        :param  access_address: Access address to use while sending PDU.
        :type   access_address: int
        :param  encrypt:        Enable PDU encryption if set to ``True``.
        :type   encrypt:        bool

        :return:                PDU transmission result.
        :rtype:                 bool
        """
        if self.__connected:
            return super().send_pdu(pdu, conn_handle, direction=direction, access_address=access_address, encrypt=encrypt)
        else:
            raise ConnectionLostException(None)

    ##############################
    # Incoming events
    ##############################

    def is_connected(self) -> bool:
        """Determine if the central device is connected to a peripheral.

        :return:    ``True`` if central is connected to a peripheral device, `False` otherwise.
        :rtype:     bool
        """
        return self.__connected

    def on_connected(self, connection_data):
        """Callback method to handle connection event.

        :param  connection_data: Connection data
        :type   connection_data: dict
        """
        # Save local and target peer info
        self.__local = BDAddress.from_bytes(
            connection_data.initiator,
            addr_type=connection_data.init_addr_type
        )

        self.__target = BDAddress.from_bytes(
            connection_data.advertiser,
            connection_data.adv_addr_type
        )

        self.__stack.on_connection(
            connection_data.conn_handle,
            self.__local,
            self.__target
        )
        self.__conn_handle = connection_data.conn_handle


    def on_disconnected(self, disconnection_data):
        """Callback method to handle disconnection event.

        :param  disconnection_data: Disconnection data
        :type   disconnection_data: :class:`whad.protocol.ble_pb2.Disconnected`
        """
        self.__stack.on_disconnection(
            disconnection_data.conn_handle,
            disconnection_data.reason
        )

        self.__connected = False
        self.__local = None
        self.__target = None

        # Notify peripheral device about this disconnection
        if self.__peripheral is not None:
            self.__peripheral.on_disconnect(disconnection_data.conn_handle)


    def on_ctl_pdu(self, pdu):
        """This callback method is called whenever a control PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        Central devices act as master, so we only forward slave to master
        messages to the stack.

        :param  pdu: BLE Control PDU
        :type   pdu: :class:`scapy.layers.bluetooth4LE.BTLE`
        """
        logger.info('received control PDU')
        if pdu.metadata.direction == Direction.SLAVE_TO_MASTER:
            self.__stack.on_ctl_pdu(pdu.metadata.connection_handle, pdu)

    def on_data_pdu(self, pdu):
        """This callback method is called whenever a data PDU is received.
        This PDU is then forwarded to the BLE stack to handle it.

        Central devices act as master, so we only forward slave to master
        messages to the stack.

        :param  pdu: BLE Data PDU
        :type   pdu: :class:`scapy.layers.bluetooth4LE.BTLE_DATA`
        """
        logger.info('received data PDU')
        if pdu.metadata.direction == Direction.SLAVE_TO_MASTER:
            self.__stack.on_data_pdu(pdu.metadata.connection_handle, pdu)


    def on_new_connection(self, connection):
        """On new connection, discover primary services.

        :param  connection: New connection Protobuf message
        :type   connection: :class:`whad.protocol.ble_pb2.Connected`
        """
        logger.info('new connection established')

        # Use GATT client
        self.connection = connection
        self.__peripheral = PeripheralDevice(
            self,
            connection.gatt,
            connection.conn_handle,
            from_json=self.__profile_json
        )

        # Retrieve GATT client
        self.__gatt_client = connection.gatt
        self.__gatt_client.set_client_model(self.__peripheral)
        self.__connected = True

        # Configure SMP layer
        # we set the security database
        self.connection.smp.set_security_database(self.__security_database)

        # Check if we got a matching LTK
        crypto_material = self.security_database.get(address=self.__target)

        if crypto_material is not None and crypto_material.has_ltk():
            self.__stack.get_layer('ll').state.register_encryption_key(
                connection.conn_handle,
                crypto_material.ltk.value
            )
            if crypto_material.is_authenticated():
                #print("Marked as authenticated")
                self.__stack.get_layer('ll').state.mark_as_authenticated(connection.conn_handle)
            else:
                pass#print("Marked as unauthenticated")

        # Notify peripheral about this connection
        self.__peripheral.on_connect(self.connection.conn_handle)


    def version(self, synchronous=True):
        """Query BLE version of remote peer.
        """
        if self.connection is not None:
            # Send an LL_VERSION_IND PDU
            self.connection.send_version()

            # Wait for an answer (mandatory)
            if synchronous:
                while not self.connection.remote_version:
                    sleep(0.01)
                result = self.connection.remote_version

                # Identify BT version
                if result.version in BT_VERSIONS:
                    version = BT_VERSIONS[result.version]
                else:
                    version = result.version

                # Identify BT company
                if result.company in BT_MANUFACTURERS:
                    company = BT_MANUFACTURERS[result.company]
                else:
                    company = result.company

                # Return information
                return (version, result.subversion, company)
            else:
                return None

    def export_profile(self):
        """Export GATT profile of the existing connection.

        :returns: Profile as a JSON string
        :rtype: str
        """
        return self.connection.gatt.model.export_json()



from scapy.layers.bluetooth import HCI_Event_LE_Meta, HCI_LE_Meta_Advertising_Reports, \
    HCI_LE_Meta_Connection_Complete, L2CAP_Hdr, HCI_Hdr, HCI_ACL_Hdr, HCI_Event_Disconnection_Complete, \
    HCI_LE_Meta_Long_Term_Key_Request, HCI_Event_Encryption_Change
from scapy.layers.bluetooth4LE import BTLE_DATA, BTLE_CTRL, LL_ENC_REQ, LL_ENC_RSP, LL_START_ENC_RSP, \
    LL_START_ENC_REQ
from whad.scapy.layers.hci import HCI_LE_Meta_Enhanced_Connection_Complete
from scapy.compat import raw
from whad.exceptions import WhadDeviceUnsupportedOperation
from whad.protocol.whad_pb2 import Message
from whad.protocol.ble.ble_pb2 import BleAdvType, BleAddrType, BleDirection
from whad.hub.ble import BDAddress, AdvType
from struct import pack, unpack
from queue import Queue
from enum import IntEnum

import logging
logger = logging.getLogger(__name__)

class HCIRole(IntEnum):
    NONE = 0
    CENTRAL = 1
    PERIPHERAL = 2

class HCIConverter:
    """
    This class converts HCI event to WHAD BLE packets and WHAD BLE packets to HCI commands.
    It allows to harmonise the representation of Bluetooth Low Energy traffic, as WHAD expects
    Link Layer-like packets.

    Obviously, some packets can't be converted easily from a representation to the other one without
    losing information or making asumptions. As a result, you should expect some limitations.
    """

    def __init__(self, device):
        self.__device = device
        self.role = HCIRole.NONE
        self.pending_messages_queue = Queue()
        self.pending_key_request = False
        self.cached_l2cap_payload = b""
        self.cached_l2cap_length = 0
        self.waiting_l2cap_fragments = False

    def process_message(self, message):
        ll_packet = BTLE_DATA(message.pdu)

        if L2CAP_Hdr in ll_packet or self.waiting_l2cap_fragments:

            if not self.waiting_l2cap_fragments and len(raw(ll_packet[L2CAP_Hdr:])) < ll_packet[L2CAP_Hdr:].len:
                self.waiting_l2cap_fragments = True
                self.cached_l2cap_payload = raw(ll_packet[BTLE_DATA:][1:])
                self.cached_l2cap_length = ll_packet[L2CAP_Hdr:].len
                return []
            elif self.waiting_l2cap_fragments:
                self.cached_l2cap_payload += raw(ll_packet[BTLE_DATA:][1:])
                if self.cached_l2cap_length == (len(self.cached_l2cap_payload) - 4):
                    L2CAP_Hdr(self.cached_l2cap_payload).show()
                    hci_packet = HCI_Hdr() / HCI_ACL_Hdr(handle = message.conn_handle) / L2CAP_Hdr(self.cached_l2cap_payload)
                    self.waiting_l2cap_fragments = False
                    return [hci_packet]
                else:
                    return []
            hci_packet = HCI_Hdr() / HCI_ACL_Hdr(handle = message.conn_handle) / ll_packet[L2CAP_Hdr:]
            return [hci_packet]
        elif ll_packet.LLID == 3:
            if LL_ENC_REQ in ll_packet:
                pdu = BTLE_DATA() / BTLE_CTRL() / LL_ENC_RSP(skds=0, ivs=0)

                direction = (BleDirection.SLAVE_TO_MASTER if
                             self.role == HCIRole.CENTRAL else
                             BleDirection.MASTER_TO_SLAVE
                )
                processed = False
                conn_handle = message.conn_handle

                msg = self.__device.hub.ble.create_pdu_received(
                    direction,
                    raw(pdu),
                    conn_handle,
                    processed
                )

                self.add_pending_message(msg)

                msg = Message()
                pdu = BTLE_DATA() / BTLE_CTRL() / LL_START_ENC_REQ()

                direction = (BleDirection.SLAVE_TO_MASTER if
                             self.role == HCIRole.CENTRAL else
                             BleDirection.MASTER_TO_SLAVE
                )
                processed = False
                conn_handle = message.conn_handle

                msg = self.__device.hub.ble.create_pdu_received(
                    direction,
                    raw(pdu),
                    conn_handle,
                    processed
                )

                self.add_pending_message(msg)

                return []
            else:
                #logger.warning("HCI devices cannot send control PDU.")
                raise WhadDeviceUnsupportedOperation("send_pdu", "Device cannot send control PDU, only data PDU.")

    def add_pending_message(self, event):
        self.pending_messages_queue.put(event)

    def get_pending_messages(self):
        events = []
        while not self.pending_messages_queue.empty():
            events.append(self.pending_messages_queue.get())
        return events

    def process_event(self, event):
        """
        Converts an HCI event to the corresponding Link Layer packet (if possible).
        """
        if HCI_Event_LE_Meta in event:
            if HCI_LE_Meta_Advertising_Reports in event:
                return self.process_advertising_reports(event[HCI_LE_Meta_Advertising_Reports:])
            elif HCI_LE_Meta_Connection_Complete in event:
                return self.process_connection_complete(event[HCI_LE_Meta_Connection_Complete:])
            elif HCI_LE_Meta_Enhanced_Connection_Complete in event:
                return self.process_connection_complete(event[HCI_LE_Meta_Enhanced_Connection_Complete:])
            elif HCI_LE_Meta_Long_Term_Key_Request in event:
                return self.process_long_term_key_request(event[HCI_LE_Meta_Long_Term_Key_Request:])
        elif HCI_Event_Encryption_Change in event:
            return self.process_encryption_change_event(event[HCI_Event_Encryption_Change:])
        elif HCI_ACL_Hdr in event:
            return self.process_acl_data(event[HCI_ACL_Hdr:])
        elif HCI_Event_Disconnection_Complete in event:
            return self.process_disconnection_complete(event[HCI_Event_Disconnection_Complete:])

    def process_encryption_change_event(self, event):
        pdu = BTLE_DATA()/BTLE_CTRL()/LL_START_ENC_RSP()

        direction = (BleDirection.SLAVE_TO_MASTER if
                     self.role == HCIRole.CENTRAL else
                     BleDirection.MASTER_TO_SLAVE
        )
        processed = False
        conn_handle = event.handle

        msg = self.__device.hub.ble.create_pdu_received(
            direction,
            raw(pdu),
            conn_handle,
            processed
        )

        return [msg]

    def process_long_term_key_request(self, event):

        msg = Message()
        pdu = BTLE_DATA()/BTLE_CTRL()/LL_ENC_REQ(
            rand = unpack('<Q', event.rand)[0],
            ediv = event.ediv,
            # SKD and IV are processed in Link Layer and are not transmitted to Host,
            # so we can't provide the real value. We arbitrarily set them to 0 to allow
            # BLE stack to continue its operations.
            skdm = 0,
            ivm = 0
        )


        direction = (BleDirection.SLAVE_TO_MASTER if
                     self.role == HCIRole.CENTRAL else
                     BleDirection.MASTER_TO_SLAVE
        )
        processed = False
        conn_handle = event.handle

        msg = self.__device.hub.ble.create_pdu_received(
            direction,
            raw(pdu),
            conn_handle,
            processed
        )

        self.pending_key_request = True
        return [msg]

    def process_acl_data(self, event):
        if event.PB == 2 and L2CAP_Hdr in event:
            msg = Message()
            pdu = BTLE_DATA()/event[L2CAP_Hdr:]

            # If HCI ACL Data PB flag==1 then it is a continued fragment.
            # We make sure BTLE_DATA.LLID is then 0x01 (Continuation
            # of a L2CAP message), our stack will take care of packet reassembly.
            if event.PB == 0x01:
                pdu.LLID = 1

            direction = (BleDirection.SLAVE_TO_MASTER if
                         self.role == HCIRole.CENTRAL else
                         BleDirection.MASTER_TO_SLAVE
            )
            processed = False
            conn_handle = event.handle

            msg = self.__device.hub.ble.create_pdu_received(
                direction,
                raw(pdu),
                conn_handle,
                processed
            )

            return [msg]

        elif event.PB == 1:
            msg = Message()
            pdu = BTLE_DATA()/event[HCI_ACL_Hdr:].payload

            # If HCI ACL Data PB flag==1 then it is a continued fragment.
            # We make sure BTLE_DATA.LLID is then 0x01 (Continuation
            # of a L2CAP message), our stack will take care of packet reassembly.
            pdu.LLID = 1

            direction = (BleDirection.SLAVE_TO_MASTER if
                         self.role == HCIRole.CENTRAL else
                         BleDirection.MASTER_TO_SLAVE
            )
            processed = False
            conn_handle = event.handle

            msg = self.__device.hub.ble.create_pdu_received(
                direction,
                raw(pdu),
                conn_handle,
                processed
            )

            return [msg]

    def process_connection_complete(self, event):
        if event.status == 0x00:

            # Mark device as connected and register handle.
            self.__device._connected = True
            self.__device._active_handles.append(event.handle)

            # Send BLE connected message to consumer.
            handle = event.handle
            if event.role == 0: # master role
                self.role = HCIRole.CENTRAL
                initiator_address = bytes.fromhex(self.__device._bd_address.replace(":",""))[::-1]
                initiator_address_type = self.__device._bd_address_type
                responder_address = bytes.fromhex(event.paddr.replace(":",""))[::-1]
                responder_address_type = BleAddrType.PUBLIC if event.patype == 0 else BleAddrType.RANDOM
            else:
                self.role = HCIRole.PERIPHERAL
                initiator_address = bytes.fromhex(event.paddr.replace(":",""))[::-1]
                initiator_address_type = BleAddrType.PUBLIC if event.patype == 0 else BleAddrType.RANDOM
                responder_address = bytes.fromhex(self.__device._bd_address.replace(":",""))[::-1]
                responder_address_type = self.__device._bd_address_type
            
            msg = self.__device.hub.ble.create_connected(
                BDAddress(initiator_address, addr_type=initiator_address_type),
                BDAddress(responder_address, addr_type=responder_address_type),
                0, # No access address
                handle
            )

            return [msg]

    def process_disconnection_complete(self, event):
        if event.status == 0x00:
            # Mark device as disconnected and remove handle
            self.__device._connected = False
            self.__device._active_handles.remove(event.handle)

            # Send disconnection message to consumer.
            msg = self.__device.hub.ble.create_disconnected(
                event.reason,
                event.handle
            )

            return [msg]

    def process_advertising_reports(self, reports):
        messages = []
        for report in reports.reports:

            adv_type = AdvType.ADV_UNKNOWN
            if report.type == 0:
                adv_type = AdvType.ADV_IND
            elif report.type == 1:
                adv_type = AdvType.ADV_DIRECT_IND
            elif report.type == 2:
                adv_type = AdvType.ADV_SCAN_IND
            elif report.type == 3:
                adv_type = AdvType.ADV_NONCONN_IND
            elif report.type == 4:
                adv_type = AdvType.ADV_SCAN_RSP

            # Flatten EIR data
            eir_data = b""
            for data in report.data:
                eir_data += raw(data)

            msg = self.__device.hub.ble.create_adv_pdu_received(
                adv_type,
                report.rssi if hasattr(report, "rssi") else 0,
                BDAddress(report.addr, random = not (report.atype == 0)),
                bytes(eir_data)
            )

            messages.append(msg)
        return messages



"""
WHAD install tool

This utility is a simple helper facilitating the install of devices.
"""
import logging
import os
import sys
import whad
import requests
import stat
import distro
import shlex
import inspect
import usb
from serial.tools.list_ports import comports
from shutil import copy, which
from prompt_toolkit import print_formatted_text, HTML
from whad.cli.ui import info, error, warning, success
from whad.cli.app import CommandLineApp, ApplicationError
from subprocess import check_output, Popen, DEVNULL, PIPE
from grp import getgrall
from pathlib import Path


logger = logging.getLogger(__name__)
#logging.basicConfig(level=logging.DEBUG)

class WhadInstallApp(CommandLineApp):

    def __init__(self):
        """Application uses no interface and has no commands.
        """
        super().__init__(
            description='WHAD install tool',
            interface=False,
            commands=False
        )


        self.add_argument(
            '--list',
            '-l',
            dest='list',
            action="store_true",
            default=False,
            help='List existing devices'
        )

        self.add_argument(
            '--rules',
            '-r',
            dest='rules',
            action="store_true",
            default=False,
            help='Install rules for selected devices'
        )


        self.add_argument(
            '--port',
            '-p',
            dest='port',
            default=False,
            help='Provide a specific port to use during flash'
        )

        self.add_argument(
            '--flash',
            '-f',
            dest='flash',
            action="store_true",
            default=False,
            help='Flash latest firmwares for selected devices'
        )

        self.add_argument(
            'device',
            nargs="*",
            default=[],
            help='Select a specific device'
        )

    def check_tools_availability(self, *tools):
        return all([which(tool) is not None for tool in tools])

    def ask_for_privileges(self):
        euid = os.geteuid()
        if euid != 0:
            info("This tool must run as root, let's elevate your privileges !")
            args = ['sudo', sys.executable] + sys.argv + [os.environ]
            # the next line replaces the currently-running process with the sudo
            pypath = ":".join(sys.path)
            args = ['sudo', f"PYTHONPATH={pypath}", sys.executable] + sys.argv
            os.execvpe('sudo', args, os.environ)
        return os.geteuid() == 0

    def get_udev_rule_filename(self, device_name):
        return os.path.realpath("{}/ressources/rules/{}.rules".format(os.path.dirname(whad.__file__), device_name))


    def install_udev_rule(self, device_name):
        matching_rules = [udev_name for udev_name in self.udev_rules if device_name in udev_name]
        if len(matching_rules) == 0:
            rule_filename = self.get_udev_rule_filename(device_name)
            rule_dest_filename = "/usr/lib/udev/rules.d/40-"+device_name+".rules"
            if os.path.exists(rule_filename):
                copy(rule_filename, rule_dest_filename)
                success("Rule successfully installed for device '%s': %s" % (device_name, rule_dest_filename))
                return True
            else:
                warning("Rule not found, skipping.")
                return False
        else:
            warning("Rules detected for device '%s' (%s), skipping." % (device_name, ",".join(matching_rules)))
            return False

    def get_latest_release(self, github_repository):
        g_request = requests.get("https://github.com/"+github_repository+"/releases/latest")
        if g_request.status_code == 200:
            return g_request.url
        else:
            return None

    def install_nrfutil(self):
        info("Downloading 'nrfutil' tool from nordicsemi.com...")
        nrfutil_url = "https://files.nordicsemi.com/ui/api/v1/download?repoKey=swtools&path=external/nrfutil/executables/x86_64-unknown-linux-gnu/nrfutil"
        n_request = requests.get(nrfutil_url, allow_redirects=True)
        if n_request.status_code == 200:
            if os.path.exists("/tmp/nrfutil"):
                os.unlink("/tmp/nrfutil")
            open("/tmp/nrfutil", "wb").write(n_request.content)
            f = Path("/tmp/nrfutil")
            f.chmod(f.stat().st_mode | stat.S_IEXEC)

            r, o, e = self.run_command('/tmp/nrfutil install nrf5sdk-tools --force')
            if len(e) > 0:
                return False
            return True

    def download_latest_butterfly_release(self):
         latest_release_url = self.get_latest_release("whad-team/butterfly")
         try:
             release = latest_release_url.split("/tag/")[1]
             success("Latest butterfly release: " + release)
             zip_url = "https://github.com/whad-team/butterfly/releases/download/" + release + "/butterfly-fwupgrade.zip"
             z_request = requests.get(zip_url, allow_redirects=True)
             if z_request.status_code == 200:
                 latest_release_filename = "/tmp/butterfly-"+release+".zip"
                 if os.path.exists(latest_release_filename):
                     os.unlink(latest_release_filename)
                 open(latest_release_filename, "wb").write(z_request.content)
                 return latest_release_filename
         except IndexError:
             error("Release not found, exiting...")
             return None

    def flash_firmware_with_nrfutil(self, firmware, serial_port="/dev/ttyACM0", nrfutil_exec="/tmp/nrfutil"):
        r, o, _  = self.run_command(nrfutil_exec + ' dfu usb-serial -pkg ' + firmware + ' -p ' + serial_port + ' -b 115200')

        return b"Device programmed" in o

    def flash_butterfly(self, port="/dev/ttyACM0"):
        if not self.install_nrfutil():
            error("Failure during 'nrfutil' installation.")
            return False

        latest_release_filename = self.download_latest_butterfly_release()
        if latest_release_filename is None:
            error("Failure during 'butterfly' release download")
            return False

        success = self.flash_firmware_with_nrfutil(latest_release_filename, serial_port=port, nrfutil_exec="/tmp/nrfutil")
        return success

    def run_command(self, cmd, shell=False, force_env=None, cwd=None, added_path=None):
        info("Running command: " + cmd)

        if added_path is not None:
            env  = os.environ.copy()
            env["PATH"] = f"{env['PATH']}:/" + added_path
        elif force_env is not None:
            env = force_env
        else:
            env = os.environ
        process = Popen(shlex.split(cmd) if not shell else cmd,  stderr=PIPE, stdout=PIPE, shell=shell, cwd=cwd, env=env)
        stdout, stderr = process.communicate()
        if process.returncode != 0:
            error("Standard output:" + stdout.decode())
            error("Standard error:" + stderr.decode())

        return process.returncode , stdout, stderr

    def install_ubertooth_host(self):
        if self.check_tools_availability('ubertooth-dfu'):
            return True

        distribution = distro.id()

        if distribution == 'fedora':
            r, _, _ = self.run_command('dnf install -y ubertooth')
            if r != 0:
                return False

        elif distribution in ('debian', 'ubuntu', 'raspbian'):
            r, _, _ = self.run_command('apt install -y ubertooth')
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")
            return False

        return True


    def download_latest_ubertooth_release(self):
         latest_release_url = self.get_latest_release("greatscottgadgets/ubertooth")
         try:
             release = latest_release_url.split("/tag/")[1]
             success("Latest ubertooth release: " + release)
             zip_url = "https://github.com/greatscottgadgets/ubertooth/releases/download/" + release + "/ubertooth-" + release + ".tar.xz"
             z_request = requests.get(zip_url, allow_redirects=True)
             if z_request.status_code == 200:
                 latest_release_filename = "/tmp/ubertooth-"+release+".tar.xz"
                 if os.path.exists(latest_release_filename):
                     os.unlink(latest_release_filename)
                 open(latest_release_filename, "wb").write(z_request.content)
                 return latest_release_filename
         except IndexError:
             error("Release not found, exiting...")
             return None

    def flash_ubertooth(self, port=None):
        if not self.install_ubertooth_host():
            error("Failure during 'ubertooth-dfu' installation.")
            return False

        latest_release_filename = self.download_latest_ubertooth_release()
        r, _, _ = self.run_command("tar -xf " + latest_release_filename, cwd="/tmp")
        if r != 0:
            return False

        r, _, _ = self.run_command("ubertooth-dfu -d bluetooth_rxtx.dfu -r" + (" -U "+str(port) if port is not None else ""), cwd=latest_release_filename.replace('.tar.xz',"/ubertooth-one-firmware-bin/"))
        if r != 0:
            return False
        return True


    def flash_yardstickone(self, port="/dev/RFCAT_BL_YS1"):
        r, _, _ = self.run_command("rm -rf /tmp/rfcat")
        if r != 0:
            return False

        r, _, _ = self.run_command("git clone https://github.com/atlas0fd00m/rfcat", cwd="/tmp")
        if r != 0:
            return False

        r, _, _ = self.run_command("pip install future")
        if r != 0:
            return False


        warning("For some reason, latest build from rfcat repo crashes the dongle, let's get the latest functional build.")
        r, _, _ = self.run_command("wget https://gist.githubusercontent.com/mossmann/7b816680df2ac513df3835f3cb9eaa1b/raw/2f33a76a86c8f6ee71dacb385537386486fe633a/RfCatYS1CCBootloader.hex -O /tmp/ys1.hex")
        if r != 0:
            return False

        if not os.path.exists(port):
            r, _, _ = self.run_command("/tmp/rfcat/rfcat --bootloader --force")
            if r != 0:
                return False
        r, _, _ = self.run_command("/tmp/rfcat/CC-Bootloader/rfcat_bootloader "+ port +" erase_all")
        if r != 0:
            return False
        r, _, _ = self.run_command("/tmp/rfcat/CC-Bootloader/rfcat_bootloader "+port+" download /tmp/ys1.hex")
        if r != 0:
            return False
        r, _, _  = self.run_command("/tmp/rfcat/CC-Bootloader/rfcat_bootloader "+port+" verify /tmp/ys1.hex")
        if r != 0:
            return False
        r, _, _ = self.run_command("/tmp/rfcat/CC-Bootloader/rfcat_bootloader "+port+" run")
        if r != 0:
            return False
        return True

    def install_sdcc(self):
        if self.check_tools_availability('sdcc'):
            return True

        distribution = distro.id()

        if distribution == 'fedora':

            r, _, _ = self.run_command('rm -rf /usr/bin/packihx /usr/bin/sdcc')
            if r != 0:
                return False

            r, _, _ = self.run_command('dnf install -y sdcc')
            if r != 0:
                return False

            r, _, _ = self.run_command('ln -s /usr/bin/sdcc-sdcc /usr/bin/sdcc')
            if r != 0:
                return False

            r, _, _ = self.run_command('ln -s /usr/bin/sdcc-packihx /usr/bin/packihx')
            if r != 0:
                return False

        elif distribution in ('debian', 'ubuntu', 'raspbian'):
            r, _, _ = self.run_command('apt install -y sdcc')
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")
            return False

        return True

    def flash_rfstorm(self):
        if not self.install_sdcc():
            error("An error occured during 'sdcc' installation.")

        r, _, _ = self.run_command("rm -rf /tmp/nrf-research-firmware-python3")
        if r != 0:
            return False
        r, _, _ = self.run_command("pip install pyusb platformio")
        if r != 0:
            return False
        r, _, _ = self.run_command("git clone https://github.com/kuzmin-no/nrf-research-firmware-python3", cwd="/tmp")
        if r != 0:
            return False
        r, _, _ = self.run_command("make", cwd="/tmp/nrf-research-firmware-python3")
        if r != 0:
            return False
        r, o, e = self.run_command("lsusb")
        if b"1915:0102" in o or b"046d:c52b" in o:
            r, _, _ = self.run_command("make logitech_install", cwd="/tmp/nrf-research-firmware-python3")
            return r == 0
        else:
            r, _, _ = self.run_command("make install", cwd="/tmp/nrf-research-firmware-python3")
            return r == 0
        return False

    def install_esp_sdk(self):
        if self.check_tools_availability('idf.py'):
            return True

        distribution = distro.id()
        if distribution == 'fedora':
            r, _, _ = self.run_command('dnf install -y wget flex bison gperf python3 cmake ninja-build ccache dfu-util libusbx')
            if r != 0:
                return False

        elif distribution in ('debian', 'ubuntu', 'raspbian'):
            r, _, _ = self.run_command('apt-get install -y git wget flex bison gperf python3 python3-pip python3-venv cmake ninja-build ccache libffi-dev libssl-dev dfu-util libusb-1.0-0')
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")
            return False

        r, _, _ = self.run_command("rm -rf /tmp/esp")
        if r != 0:
            return False
        r, _, _ = self.run_command("mkdir -p /tmp/esp")
        if r != 0:
            return False

        r, _, _ = self.run_command("git clone -b release/v4.4 --recursive https://github.com/espressif/esp-idf.git", cwd="/tmp/esp")
        if r != 0:
            return False

        # Clear python path
        env = os.environ.copy()
        env["PYTHONPATH"] = ""

        r, _, _ = self.run_command("./install.sh esp32", shell=True, force_env=env, cwd="/tmp/esp/esp-idf")
        if r != 0:
            return False

        r, o, e = self.run_command(". ./export.sh > /dev/null; env", shell=True, force_env=env, cwd="/tmp/esp/esp-idf")
        if r != 0:
            return False

        os.environ.update(dict(line.decode().partition('=   ')[::2] for line in o.split(b'\n')[:-1]))

        r, o, e = self.run_command("idf.py --version", force_env=os.environ,  cwd="/tmp/esp/esp-idf")
        if r != 0:
            return False

        success("Espressif framework installed: " + o.decode())
        return True


    def flash_esp(self, port=None):
        if not self.install_esp_sdk():
            error("An error occured uring ESP SDK installation.")
            return False

        info("Cloning 'nodemcu-esp32-firmware' github repository...")
        r, _, _ = self.run_command("rm -rf /tmp/nodemcu-esp32-firmware/", force_env=os.environ,  cwd="/tmp")
        if r != 0:
            return False

        r, _, _ = self.run_command("git clone https://github.com/whad-team/nodemcu-esp32-firmware/ --recurse", force_env=os.environ,  cwd="/tmp")
        if r != 0:
            return False

        info("Building firmware...")
        r, _, _ = self.run_command("idf.py build", force_env=os.environ,  cwd="/tmp/nodemcu-esp32-firmware")
        if r != 0:
            return False

        info("Flashing firmware...")
        r, o, e = self.run_command("idf.py flash" + ("" if port is None else " -p " + port), force_env=os.environ,  cwd="/tmp/nodemcu-esp32-firmware")
        if r != 0:
            return False

        return True


    def download_latest_stm32wl55_release(self):
         latest_release_url = self.get_latest_release("whad-team/stm32wlxx-firmware")
         try:
             release = latest_release_url.split("/tag/")[1]
             success("Latest STM32WL55 firmware release: " + release)
             zip_url = "https://github.com/whad-team/stm32wlxx-firmware/releases/download/" + release + "/nucleo_wl55.hex"
             z_request = requests.get(zip_url, allow_redirects=True)
             if z_request.status_code == 200:
                 latest_release_filename = "/tmp/nucleo_wl55.hex"
                 if os.path.exists(latest_release_filename):
                     os.unlink(latest_release_filename)
                 open(latest_release_filename, "wb").write(z_request.content)
                 return latest_release_filename
         except IndexError:
             error("Release not found, exiting...")
             return None

    def download_latest_lorae5mini_release(self):
         latest_release_url = self.get_latest_release("whad-team/stm32wlxx-firmware")
         try:
             release = latest_release_url.split("/tag/")[1]
             success("Latest STM32WL55 firmware release: " + release)
             zip_url = "https://github.com/whad-team/stm32wlxx-firmware/releases/download/" + release + "/lora_e5_mini.hex"
             z_request = requests.get(zip_url, allow_redirects=True)
             if z_request.status_code == 200:
                 latest_release_filename = "/tmp/lora_e5_mini.hex"
                 if os.path.exists(latest_release_filename):
                     os.unlink(latest_release_filename)
                 open(latest_release_filename, "wb").write(z_request.content)
                 return latest_release_filename
         except IndexError:
             error("Release not found, exiting...")
             return None

    def install_stflash(self):
        if self.check_tools_availability('st-flash'):
            return True

        distribution = distro.id()
        if distribution == 'fedora':
            r, _, _ = self.run_command('dnf install -y stlink')
            if r != 0:
                return False

        elif distribution in ('debian', 'ubuntu', 'raspbian'):
            r, _, _ = self.run_command('apt-get install -y stlink-tools')
            if r != 0:
                return False

        else:
            error("Platform not supported, exiting.")
            return False



    def flash_nucleowl55(self):
        if not self.install_stflash():
            error("An error occured during 'st-flash' installation.")
            return False


        latest_release_filename = self.download_latest_stm32wl55_release()
        r, _, _ = self.run_command("st-flash --format=ihex write " + latest_release_filename, cwd="/tmp")
        if r != 0:
            return False

        return True



    def flash_lorae5mini(self):
        if not self.install_stflash():
            error("An error occured during 'st-flash' installation.")
            return False


        latest_release_filename = self.download_latest_lorae5mini_release()
        r, _, _ = self.run_command("st-flash --format=ihex write " + latest_release_filename, cwd="/tmp")
        if r != 0:
            return False

        return True

    def rules_rfstorm(self):
        if self.install_udev_rule("rfstorm"):
            self.need_reload = True
        return True

    def rules_ubertooth(self):
        if self.install_udev_rule("ubertooth"):
            self.need_reload = True
        return True

    def rules_rzusbstick(self):
        if self.install_udev_rule("rzusbstick"):
            self.need_reload = True
        return True

    def rules_yardstickone(self):
        if self.install_udev_rule("yardstickone"):
            self.need_reload = True
        return True

    def rules_nucleowl55(self):
        return self.install_serial_port_capabilities()

    def rules_lorae5mini(self):
        return self.install_serial_port_capabilities()

    def rules_esp(self):
        return self.install_serial_port_capabilities()

    def rules_butterfly(self):
        return self.install_serial_port_capabilities()

    def rules_apimote(self):
        return self.install_serial_port_capabilities()

    def rules_hci(self):
        python_interpreter = os.path.realpath(sys.executable)
        if not self.install_hci_capabilities(python_interpreter):
            return False
        success("HCI capabilities successfully added to python interpreter (%s)." % python_interpreter)
        return True

    def pre_run(self):
        super().pre_run()
        # find udev location
        candidate_udev_locations = [
            "/etc/udev/rules.d",
            "/lib/udev/rules.d",
            "/run/udev/rules.d",
            "/var/run/udev/rules.d"
        ]
        self.udev_rules = []

        for location in candidate_udev_locations:
            if os.path.exists(location):
                self.udev_rules += os.listdir(location)

        self.need_reload = False


    def install_serial_port_capabilities(self):
        if not self.check_tools_availability("usermod"):
            error("'usermod' tool is required but not available")
            return False
        # Find the right group which depends on the distribution (e.g. dialout
        # for Debian, uucp for Arch-Linux).
        groups = [ group.gr_name for group in getgrall() ]
        if "dialout" in groups:
            group = "dialout"
        elif "uucp" in groups:
            group = "uucp"
        else:
            error("Group granting serial port capabilities not found!")
            return False
        process = Popen(['usermod', '-a', '-G', group, os.getlogin()], stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        if len(stderr) > 0:
            return False

        success("Serial port capabilities successfully configured (user %s added to group %s): logout to take modification into account." % (os.getlogin(), group))
        return True


    def reload_udev_rules(self):
        if not self.check_tools_availability("udevadm"):
            error("'udevadm' tool is required but not available")
            return False
        info("Reloading udev rules...")
        process = Popen(['udevadm', 'control', '--reload-rules'], stderr=PIPE, stdout=PIPE)
        stdout, stderr = process.communicate()
        if len(stderr) > 0:
            return False
        process = Popen(['udevadm', 'trigger'],  stderr=PIPE, stdout=PIPE)
        stdout, stderr = process.communicate()
        if len(stderr) > 0:
            return False
        return True


    def install_hci_capabilities(self, interpreter):
        if not self.check_tools_availability("getcap", "setcap"):
            error("'getcap' and 'setcap' tools are required but not available")
            return False
        process = Popen(['getcap', interpreter], stdout=PIPE, stderr=PIPE)
        stdout, stderr = process.communicate()

        if "cap_net_admin,cap_net_raw=eip" not in stdout.decode('utf-8'):
            process = Popen(['setcap', 'cap_net_admin,cap_net_raw+eip',interpreter], stdout=PIPE, stderr=PIPE)
            stdout, stderr = process.communicate()

            if len(stdout) != 0 or len(stderr) != 0:
                return False
            return True
        else:
            return True

    def list_devices(self):
        info("Detected devices:")
        print()
        for dev in comports():
            if (dev.vid == 0x10c4 and dev.pid == 0xea60) or dev.vid == 0x303a:
                print_formatted_text(HTML("  - <b>Espressif ESP-32 board: </b> {}".format(dev.device)))
                print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules esp</i>"))
                print_formatted_text(HTML("    <u>Command (flash firmware):</u> <i>winstall --flash esp --port {}</i>".format(dev.device)))
                print()
            elif dev.vid == 0xc0ff and dev.pid == 0xeeee:
                print_formatted_text(HTML("  - <b>WHAD ButteRFly dongle: </b> {}".format(dev.device)))
                print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules butterfly</i>"))
                print_formatted_text(HTML("    <u>Command (flash firmware):</u> <i>winstall --flash butterfly --port {}</i>".format(dev.device)))
                print()
            elif dev.vid == 0x0483 and dev.pid == 0x374e:
                print_formatted_text(HTML("  - <b>Nucleo STM32WL55: </b> {}".format(dev.device)))
                print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules nucleowl55</i>"))
                print_formatted_text(HTML("    <u>Command (flash firmware):</u> <i>winstall --flash nucleowl55 --port {}</i>".format(dev.device)))
                print()

        ubertooth_count = 0
        yardstickone_count = 0
        rfstorm_count = 0
        rzusbstick_count = 0
        for device in usb.core.find(find_all=1):
            vid, pid = device.idVendor, device.idProduct
            if (vid == 0xffff and pid == 0x0004) or (vid == 0x1d50 and pid >= 0x6000 and pid <= 0x6003):
                print_formatted_text(HTML("  - <b>Ubertooth One: </b> {}".format(ubertooth_count)))
                print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules ubertooth </i>"))
                print_formatted_text(HTML("    <u>Command (flash firmware):</u> <i>winstall --flash ubertooth --port {}</i>".format(ubertooth_count)))
                ubertooth_count += 1
                print()
            elif (vid == 0x1d50 and pid in (0x605b, 0x6047, 0x6048, 0xecc1, 0x6049, 0x604a, 0x605c, 0xecc0)):
                print_formatted_text(HTML("  - <b>Yard Stick One: </b> {}".format(yardstickone_count)))
                print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules yardstickone </i>"))
                print_formatted_text(HTML("    <u>Command (flash firmware):</u> <i>winstall --flash yardstickone --port {}</i>".format(yardstickone_count)))
                yardstickone_count += 1
                print()
            elif (vid == 0x046d and pid in (0xc52b, 0xaaaa)) or (vid == 0x1915 and pid in (0x0102, 0x7777, 0x0101)):
                print_formatted_text(HTML("  - <b>RFStorm compatible dongle: </b> {}".format(rfstorm_count)))
                print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules rfstorm </i>"))
                print_formatted_text(HTML("    <u>Command (flash firmware):</u> <i>winstall --flash rfstorm --port {}</i>".format(rfstorm_count)))
                rfstorm_count += 1
                print()
            elif (vid == 0x03eb and pid == 0x210a):
                print_formatted_text(HTML("  - <b>RZUSBStick: </b> {}".format(rzusbstick_count)))
                print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules rzusbstick </i>"))
                rzusbstick_count += 1
                print()

        for device in os.listdir("/sys/class/bluetooth"):
            print_formatted_text(HTML("  - <b>HCI device: </b> {}".format(device)))
            print_formatted_text(HTML("    <u>Command (install rules):</u> <i>winstall --rules hci </i>"))
            print()

    def run(self):
        try:
            # Launch pre-run tasks
            self.pre_run()

            rules = self.args.rules
            flash = self.args.flash

            if not self.args.rules and not self.args.flash:
                # By default, if no option is provided, only use rules
                self.args.list = True

            if self.args.list or len(self.args.device) == 0:
                self.list_devices()
                return

            # We need root, let's ask
            if not self.ask_for_privileges():
                error("Login failed, aborting.")
                exit(1)

            if  "all" in self.args.device:
                self.args.device = [
                    "hci",
                    "butterfly",
                    "esp",
                    "lorae5mini",
                    "rfstorm",
                    "ubertooth",
                    "apimote",
                    "nucleowl55",
                    "rzusbstick",
                    "yardstickone"
                ]

            # Iterate over devices
            for device in self.args.device:
                if flash:
                    if hasattr(self, "flash_" + device):
                        info("Flashing {} device ...".format(device))

                        flashfunc = getattr(self, "flash_" + device)
                        if self.args.port and "port" in list(inspect.signature(flashfunc).parameters):
                            if flashfunc(port=self.args.port):
                                success("Flashing successful for device '%s'." % device)
                            else:
                                error("An error occured during device flashing for device '%s'." % device)
                                exit(1)
                        else:
                            if flashfunc():
                                success("Flashing successful for device '%s'." % device)
                            else:
                                error("An error occured during device flashing for device '%s'." % device)
                                exit(1)
                    else:
                        warning("Flashing {} device is not supported yet.".format(device))

                if rules:
                    if hasattr(self, "rules_" + device):
                        info("Installing rules for {} device ...".format(device))
                        if getattr(self, "rules_" + device)():
                            success("Rules successfully added for device '%s'." % device)
                        else:
                            error("An error occured during device rules installation for device '%s'." % device)
                            exit(1)
                    else:
                        warning("Installing rules for {} device is not supported yet.".format(device))


            if rules and self.need_reload:
                if not self.reload_udev_rules():
                    error("An error occured during udev rules reload, exiting.")
                    exit(1)
                else:
                    success("Rules successfully reloaded.")
        except Exception as e:
            error("An error occured: " + repr(e))

        # Launch post-run tasks
        self.post_run()

def winstall_main():
    try:
        app = WhadInstallApp()
        app.run()
    except ApplicationError as err:
        err.show()



import re, json
from binascii import hexlify, unhexlify

class InvalidBDAddressException(Exception):
    """Invalid BD address used
    """
    def __init__(self):
        super().__init__()

class BDAddress(object):
    """This class represents a Bluetooth Device address.
    """

    PUBLIC = 0x00
    RANDOM = 0x01

    def __init__(self, address, random=False, addr_type=None):
        """Initialize BD address

        By default, BD address is public unless `random` is set to True.

        :param int addr_type: Set BD address type (either BDAddress.PUBLIC or BDAddress.RANDOM). This setting overseeds `random`
        :param bool random: Set BD address as random if set to True. BD address is public by default.
        """
        if addr_type is not None:
            if addr_type in [BDAddress.PUBLIC, BDAddress.RANDOM]:
                self.__type = addr_type
        elif random:
            self.__type = BDAddress.RANDOM
        else:
            self.__type = BDAddress.PUBLIC

        if isinstance(address, str):
            if re.match(r'^([0-9a-fA-F]{2}\:){5}[0-9a-fA-F]{2}$', address) is not None:
                self.__value = unhexlify(address.replace(':',''))[::-1]
            elif re.match('[0-9a-fA-F]{12}$', address) is not None:
                self.__value = unhexlify(address)[::-1]
            else:
                raise InvalidBDAddressException
        elif isinstance(address, bytes) and len(address) == 6:
            self.__value = address
        else:
            raise InvalidBDAddressException

    def __eq__(self, other):
        return (self.__value == other.value) and (self.__type == other.type)

    def __str__(self):
        return ':'.join(['%02x' % b for b in self.__value[::-1]])

    def __repr__(self):
        return 'BDAddress(%s)' % str(self)

    def export_json(self):
        return json.dumps(str(self))

    @property
    def type(self):
        return self.__type

    @property
    def value(self):
        return self.__value

    def is_public(self):
        """Determine if address is public.

        :rtype: bool
        :return: True if BD address is public, False otherwise.
        """
        return (self.__type == BDAddress.PUBLIC)

    def is_random(self):
        """Determine if address is random.

        :rtype: bool
        :return: True if BD address is random, False otherwise.
        """
        return (self.__type == BDAddress.RANDOM)

    @staticmethod
    def from_bytes(bd_addr_bytes, addr_type=None):
        """Convert a 6-byte array into a valid BD address.

        :param bytes bd_addr_bytes: Bluetooth Device address as a bytearray.
        :param addr_type: Bluetooth Device Address type
        :type addr_type: int, optional
        :rtype: BDAddress
        :returns: An instance of BDAddress representing the corresponding BD address.
        """
        if len(bd_addr_bytes) == 6:
            hex_address = hexlify(bd_addr_bytes[::-1])
            address = b':'.join([hex_address[i*2:(i+1)*2] for i in range(int(len(hex_address)/2))])
            return BDAddress(address.decode('utf-8'), addr_type=addr_type)
        else:
            raise InvalidBDAddressException

    @staticmethod

    def check(bd_addr: str) -> bool:
        """Check if a BD address is valid

        :param      bd_addr: BD address to test
        :type       bd_addr: str
        :return:    ``True`` if BD address is valid, ``False`` otherwise
        :rtype:     bool
        """
        return re.match('^([0-9a-fA-F]{2}:){5}[0-9a-fA-F]{2}$',bd_addr)



"""WHAD Protocol hub Registry

This module implements the Registry class used in automated protocol buffers
messages parsing.
"""
from .exceptions import UnsupportedVersionException

class Registry(object):
    """Protocol message versions registry.
    """

    VERSIONS = {}

    @classmethod
    def add_node_version(parent_class, version: int, name: str, clazz):
        """Add a specific class `clazz` to our message registry for version
        `version` with alias `name`.
        """
        # If version is unknown, create it
        if version not in parent_class.VERSIONS:
            parent_class.VERSIONS[version] = {}

        # Add clazz based on provided alias for this version
        parent_class.VERSIONS[version][name] = clazz

    @classmethod
    def bound(parent_class, name: str = None, version: int = 1):
        """Retrieve the given node class `name` for version `version`.

        If there is no defined class for version N, look for a corresponding
        class in version N-1, N-2 until 0.
        """
        # Look for node class from given name and version
        if version in parent_class.VERSIONS:
            if name in parent_class.VERSIONS[version]:
                return parent_class.VERSIONS[version][name]

        if version > 1:
            # If not found for version N, look for node class in version N-1
            return parent_class.bound(name, version - 1)

        # If not found, raise exception
        raise UnsupportedVersionException(name, version)



from whad.phy import PhysicalLayer, GFSKModulationScheme, Endianness
from whad.scapy.layers.esb import ESB_Hdr
from whad.esb.esbaddr import ESBAddress
from enum import IntEnum

class FieldsSize(IntEnum):
    '''
    Size of major ESB fields (in bytes).
    '''
    PREAMBLE_SIZE = 1
    ADDRESS_SIZE = 5
    HEADER_SIZE = 2
    CRC_SIZE = 2


def frequency_to_channel(frequency):
  """
  Converts a frequency (in Hz) to an Enhanced ShockBurst channel.
  """
  return int(frequency/1000000) - 2400

def channel_to_frequency(channel):
  """
  Converts an Enhanced ShockBurst channel into a frequency (in Hz).
  """
  return (2400 + channel) * 1000000

def decoding(pkt, configuration):
    size = ((pkt[6] & 0b11111100) >> 2)
    return pkt[:FieldsSize.PREAMBLE_SIZE + FieldsSize.CRC_SIZE + FieldsSize.ADDRESS_SIZE+FieldsSize.HEADER_SIZE+size]


PHYS = {
    "ESB-1M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=170000),
                datarate=1000000,
                endianness=Endianness.BIG,
                frequency_range=(2400000000, 2500000000),
                maximum_packet_size=31,
                synchronization_word=b"\xAA",
                scapy_layer=ESB_Hdr,
                format_address_function=lambda address:bytes.fromhex(address.replace(":","")),
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                encoding_function=None,
                decoding_function=decoding
            ),

    "ESB-2M": PhysicalLayer(
                modulation=GFSKModulationScheme(deviation=500000),
                datarate=2000000,
                endianness=Endianness.BIG,
                frequency_range=(2400000000, 2500000000),
                maximum_packet_size=31,
                synchronization_word=b"\xAA",
                scapy_layer=ESB_Hdr,
                format_address_function=lambda address:bytes.fromhex(address.replace(":","")),
                frequency_to_channel_function=frequency_to_channel,
                channel_to_frequency_function=channel_to_frequency,
                encoding_function=None,
                decoding_function=decoding
            )
}
